from typing import Annotated

from eth_rpc.types import Name, Struct
from pydantic import BaseModel


class A(Struct):
    x: int
    y: str
    z: list[bool]


class B(Struct):
    a: list[A]
    b: A
    c: Annotated[bool, Name("BOOL")]


class C(Struct):
    a: Annotated[tuple[A, B, list[B]], Name("tuples")]
    d: list[B]
    e: str


class D(BaseModel):
    a: Annotated[tuple[A, B, list[B]], Name("tuples")]
    d: list[B]
    e: str


def test_typing_nested_structs():
    A_str = "(uint256,string,bool[])"
    B_str = f"({A_str}[],{A_str},bool)"
    C_str = f"(({A_str},{B_str},{B_str}[]),{B_str}[],string)"

    assert C.to_tuple_type() == C_str


def test_load_struct():
    a1 = A(x=1, y="a", z=[True, False])
    a2 = A(x=2, y="b", z=[False, True])
    a3 = A(x=3, y="b", z=[False, True])
    a4 = A(x=4, y="b", z=[False, True])
    b1 = B(a=[a1, a2], b=a3, c=True)
    b2 = B(a=[a1, a2], b=a3, c=False)
    c = C(a=(a4, b1, [b2]), d=[b1.model_copy(), b2.model_copy()], e="final")

    assert A.from_bytes(a1.to_bytes()) == a1
    assert A.from_bytes(a2.to_bytes()) == a2
    assert A.from_bytes(a3.to_bytes()) == a3
    assert A.from_bytes(a4.to_bytes()) == a4

    assert B.from_bytes(b1.to_bytes()) == b1
    assert B.from_bytes(b2.to_bytes()) == b2

    assert C.from_bytes(c.to_bytes()) == c



import ast

from eth_rpc.codegen import codegen

UNISWAP_UNIVERAL_ROUTER_ABI = [
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "permit2", "type": "address"},
                    {"internalType": "address", "name": "weth9", "type": "address"},
                    {"internalType": "address", "name": "seaport", "type": "address"},
                    {"internalType": "address", "name": "nftxZap", "type": "address"},
                    {"internalType": "address", "name": "x2y2", "type": "address"},
                    {
                        "internalType": "address",
                        "name": "foundation",
                        "type": "address",
                    },
                    {"internalType": "address", "name": "sudoswap", "type": "address"},
                    {"internalType": "address", "name": "nft20Zap", "type": "address"},
                    {
                        "internalType": "address",
                        "name": "cryptopunks",
                        "type": "address",
                    },
                    {"internalType": "address", "name": "looksRare", "type": "address"},
                    {
                        "internalType": "address",
                        "name": "routerRewardsDistributor",
                        "type": "address",
                    },
                    {
                        "internalType": "address",
                        "name": "looksRareRewardsDistributor",
                        "type": "address",
                    },
                    {
                        "internalType": "address",
                        "name": "looksRareToken",
                        "type": "address",
                    },
                    {"internalType": "address", "name": "v2Factory", "type": "address"},
                    {"internalType": "address", "name": "v3Factory", "type": "address"},
                    {
                        "internalType": "bytes32",
                        "name": "pairInitCodeHash",
                        "type": "bytes32",
                    },
                    {
                        "internalType": "bytes32",
                        "name": "poolInitCodeHash",
                        "type": "bytes32",
                    },
                ],
                "internalType": "struct RouterParameters",
                "name": "params",
                "type": "tuple",
            }
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {"inputs": [], "name": "ContractLocked", "type": "error"},
    {"inputs": [], "name": "ETHNotAccepted", "type": "error"},
    {
        "inputs": [
            {"internalType": "uint256", "name": "commandIndex", "type": "uint256"},
            {"internalType": "bytes", "name": "message", "type": "bytes"},
        ],
        "name": "ExecutionFailed",
        "type": "error",
    },
    {"inputs": [], "name": "FromAddressIsNotOwner", "type": "error"},
    {"inputs": [], "name": "InsufficientETH", "type": "error"},
    {"inputs": [], "name": "InsufficientToken", "type": "error"},
    {"inputs": [], "name": "InvalidBips", "type": "error"},
    {
        "inputs": [
            {"internalType": "uint256", "name": "commandType", "type": "uint256"}
        ],
        "name": "InvalidCommandType",
        "type": "error",
    },
    {"inputs": [], "name": "InvalidOwnerERC1155", "type": "error"},
    {"inputs": [], "name": "InvalidOwnerERC721", "type": "error"},
    {"inputs": [], "name": "InvalidPath", "type": "error"},
    {"inputs": [], "name": "InvalidReserves", "type": "error"},
    {"inputs": [], "name": "LengthMismatch", "type": "error"},
    {"inputs": [], "name": "NoSlice", "type": "error"},
    {"inputs": [], "name": "SliceOutOfBounds", "type": "error"},
    {"inputs": [], "name": "SliceOverflow", "type": "error"},
    {"inputs": [], "name": "ToAddressOutOfBounds", "type": "error"},
    {"inputs": [], "name": "ToAddressOverflow", "type": "error"},
    {"inputs": [], "name": "ToUint24OutOfBounds", "type": "error"},
    {"inputs": [], "name": "ToUint24Overflow", "type": "error"},
    {"inputs": [], "name": "TransactionDeadlinePassed", "type": "error"},
    {"inputs": [], "name": "UnableToClaim", "type": "error"},
    {"inputs": [], "name": "UnsafeCast", "type": "error"},
    {"inputs": [], "name": "V2InvalidPath", "type": "error"},
    {"inputs": [], "name": "V2TooLittleReceived", "type": "error"},
    {"inputs": [], "name": "V2TooMuchRequested", "type": "error"},
    {"inputs": [], "name": "V3InvalidAmountOut", "type": "error"},
    {"inputs": [], "name": "V3InvalidCaller", "type": "error"},
    {"inputs": [], "name": "V3InvalidSwap", "type": "error"},
    {"inputs": [], "name": "V3TooLittleReceived", "type": "error"},
    {"inputs": [], "name": "V3TooMuchRequested", "type": "error"},
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            }
        ],
        "name": "RewardsSent",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "looksRareClaim", "type": "bytes"}
        ],
        "name": "collectRewards",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "commands", "type": "bytes"},
            {"internalType": "bytes[]", "name": "inputs", "type": "bytes[]"},
        ],
        "name": "execute",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "commands", "type": "bytes"},
            {"internalType": "bytes[]", "name": "inputs", "type": "bytes[]"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
        ],
        "name": "execute",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
            {"internalType": "bytes", "name": "", "type": "bytes"},
        ],
        "name": "onERC1155BatchReceived",
        "outputs": [{"internalType": "bytes4", "name": "", "type": "bytes4"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "bytes", "name": "", "type": "bytes"},
        ],
        "name": "onERC1155Received",
        "outputs": [{"internalType": "bytes4", "name": "", "type": "bytes4"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "bytes", "name": "", "type": "bytes"},
        ],
        "name": "onERC721Received",
        "outputs": [{"internalType": "bytes4", "name": "", "type": "bytes4"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes4", "name": "interfaceId", "type": "bytes4"}],
        "name": "supportsInterface",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "int256", "name": "amount0Delta", "type": "int256"},
            {"internalType": "int256", "name": "amount1Delta", "type": "int256"},
            {"internalType": "bytes", "name": "data", "type": "bytes"},
        ],
        "name": "uniswapV3SwapCallback",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {"stateMutability": "payable", "type": "receive"},
]
UNISWAP_SWAP_ROUTER = [
    {
        "inputs": [
            {"internalType": "address", "name": "_factory", "type": "address"},
            {"internalType": "address", "name": "_WETH9", "type": "address"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "inputs": [],
        "name": "WETH9",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes", "name": "path", "type": "bytes"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                    {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountOutMinimum",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct ISwapRouter.ExactInputParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactInput",
        "outputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"}
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "tokenIn", "type": "address"},
                    {"internalType": "address", "name": "tokenOut", "type": "address"},
                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                    {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountOutMinimum",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint160",
                        "name": "sqrtPriceLimitX96",
                        "type": "uint160",
                    },
                ],
                "internalType": "struct ISwapRouter.ExactInputSingleParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactInputSingle",
        "outputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"}
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes", "name": "path", "type": "bytes"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                    {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountInMaximum",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct ISwapRouter.ExactOutputParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactOutput",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "tokenIn", "type": "address"},
                    {"internalType": "address", "name": "tokenOut", "type": "address"},
                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                    {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountInMaximum",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint160",
                        "name": "sqrtPriceLimitX96",
                        "type": "uint160",
                    },
                ],
                "internalType": "struct ISwapRouter.ExactOutputSingleParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactOutputSingle",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "factory",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes[]", "name": "data", "type": "bytes[]"}],
        "name": "multicall",
        "outputs": [{"internalType": "bytes[]", "name": "results", "type": "bytes[]"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "refundETH",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermit",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "nonce", "type": "uint256"},
            {"internalType": "uint256", "name": "expiry", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitAllowed",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "nonce", "type": "uint256"},
            {"internalType": "uint256", "name": "expiry", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitAllowedIfNecessary",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitIfNecessary",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
        ],
        "name": "sweepToken",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "sweepTokenWithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "int256", "name": "amount0Delta", "type": "int256"},
            {"internalType": "int256", "name": "amount1Delta", "type": "int256"},
            {"internalType": "bytes", "name": "_data", "type": "bytes"},
        ],
        "name": "uniswapV3SwapCallback",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
        ],
        "name": "unwrapWETH9",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "unwrapWETH9WithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {"stateMutability": "payable", "type": "receive"},
]


def test_codegen():
    result = codegen(UNISWAP_SWAP_ROUTER, "UniversalRouter")
    ast.parse(result)

    result = codegen(UNISWAP_UNIVERAL_ROUTER_ABI, "UniversalRouter")
    ast.parse(result)



from typing import NewType, Union

from eth_typing import ChecksumAddress, HexAddress, HexStr

int8 = NewType("int8", int)
int16 = NewType("int16", int)
int24 = NewType("int24", int)
int32 = NewType("int32", int)
int40 = NewType("int40", int)
int48 = NewType("int48", int)
int56 = NewType("int56", int)
int64 = NewType("int64", int)
int72 = NewType("int72", int)
int80 = NewType("int80", int)
int88 = NewType("int88", int)
int96 = NewType("int96", int)
int104 = NewType("int104", int)
int112 = NewType("int112", int)
int120 = NewType("int120", int)
int128 = NewType("int128", int)
int136 = NewType("int136", int)
int144 = NewType("int144", int)
int152 = NewType("int152", int)
int160 = NewType("int160", int)
int168 = NewType("int168", int)
int176 = NewType("int176", int)
int184 = NewType("int184", int)
int192 = NewType("int192", int)
int200 = NewType("int200", int)
int208 = NewType("int208", int)
int216 = NewType("int216", int)
int224 = NewType("int224", int)
int232 = NewType("int232", int)
int240 = NewType("int240", int)
int248 = NewType("int248", int)
int256 = NewType("int256", int)

uint8 = NewType("uint8", int)
uint16 = NewType("uint16", int)
uint24 = NewType("uint24", int)
uint32 = NewType("uint32", int)
uint40 = NewType("uint40", int)
uint48 = NewType("uint48", int)
uint56 = NewType("uint56", int)
uint64 = NewType("uint64", int)
uint72 = NewType("uint72", int)
uint80 = NewType("uint80", int)
uint88 = NewType("uint88", int)
uint96 = NewType("uint96", int)
uint104 = NewType("uint104", int)
uint112 = NewType("uint112", int)
uint120 = NewType("uint120", int)
uint128 = NewType("uint128", int)
uint136 = NewType("uint136", int)
uint144 = NewType("uint144", int)
uint152 = NewType("uint152", int)
uint160 = NewType("uint160", int)
uint168 = NewType("uint168", int)
uint176 = NewType("uint176", int)
uint184 = NewType("uint184", int)
uint192 = NewType("uint192", int)
uint200 = NewType("uint200", int)
uint208 = NewType("uint208", int)
uint216 = NewType("uint216", int)
uint224 = NewType("uint224", int)
uint232 = NewType("uint232", int)
uint240 = NewType("uint240", int)
uint248 = NewType("uint248", int)
uint256 = NewType("uint256", int)


bytes1 = NewType("bytes1", bytes)
bytes2 = NewType("bytes2", bytes)
bytes3 = NewType("bytes3", bytes)
bytes4 = NewType("bytes4", bytes)
bytes5 = NewType("bytes5", bytes)
bytes6 = NewType("bytes6", bytes)
bytes7 = NewType("bytes7", bytes)
bytes8 = NewType("bytes8", bytes)
bytes9 = NewType("bytes9", bytes)
bytes10 = NewType("bytes10", bytes)
bytes11 = NewType("bytes11", bytes)
bytes12 = NewType("bytes12", bytes)
bytes13 = NewType("bytes13", bytes)
bytes14 = NewType("bytes14", bytes)
bytes15 = NewType("bytes15", bytes)
bytes16 = NewType("bytes16", bytes)
bytes17 = NewType("bytes17", bytes)
bytes18 = NewType("bytes18", bytes)
bytes19 = NewType("bytes19", bytes)
bytes20 = NewType("bytes20", bytes)
bytes21 = NewType("bytes21", bytes)
bytes22 = NewType("bytes22", bytes)
bytes23 = NewType("bytes23", bytes)
bytes24 = NewType("bytes24", bytes)
bytes25 = NewType("bytes25", bytes)
bytes26 = NewType("bytes26", bytes)
bytes27 = NewType("bytes27", bytes)
bytes28 = NewType("bytes28", bytes)
bytes29 = NewType("bytes29", bytes)
bytes30 = NewType("bytes30", bytes)
bytes31 = NewType("bytes31", bytes)
bytes32 = NewType("bytes32", bytes)

string = NewType("string", str)
address = NewType("address", str)

# redundant, but facilitates code consistency, ie. tuple[primitives.bool, primitives.address]
bool = bool
bytes = bytes

BYTES_TYPES = Union[
    type[bytes1],
    type[bytes2],
    type[bytes3],
    type[bytes4],
    type[bytes5],
    type[bytes6],
    type[bytes7],
    type[bytes8],
    type[bytes9],
    type[bytes10],
    type[bytes11],
    type[bytes12],
    type[bytes13],
    type[bytes14],
    type[bytes15],
    type[bytes16],
    type[bytes17],
    type[bytes18],
    type[bytes19],
    type[bytes20],
    type[bytes21],
    type[bytes22],
    type[bytes23],
    type[bytes24],
    type[bytes25],
    type[bytes26],
    type[bytes27],
    type[bytes28],
    type[bytes29],
    type[bytes30],
    type[bytes31],
    type[bytes32],
    type[bytes],
]

BASIC_TYPES = Union[
    int8,
    int16,
    int24,
    int32,
    int40,
    int48,
    int56,
    int64,
    int72,
    int80,
    int88,
    int96,
    int104,
    int112,
    int120,
    int128,
    int136,
    int144,
    int152,
    int160,
    int168,
    int176,
    int184,
    int192,
    int200,
    int208,
    int216,
    int224,
    int232,
    int240,
    int248,
    int256,
    uint8,
    uint16,
    uint24,
    uint32,
    uint40,
    uint48,
    uint56,
    uint64,
    uint72,
    uint80,
    uint88,
    uint96,
    uint104,
    uint112,
    uint120,
    uint128,
    uint136,
    uint144,
    uint152,
    uint160,
    uint168,
    uint176,
    uint184,
    uint192,
    uint200,
    uint208,
    uint216,
    uint224,
    uint232,
    uint240,
    uint248,
    uint256,
    bytes1,
    bytes2,
    bytes3,
    bytes4,
    bytes5,
    bytes6,
    bytes7,
    bytes8,
    bytes9,
    bytes10,
    bytes11,
    bytes12,
    bytes13,
    bytes14,
    bytes15,
    bytes16,
    bytes17,
    bytes18,
    bytes19,
    bytes20,
    bytes21,
    bytes22,
    bytes23,
    bytes24,
    bytes25,
    bytes26,
    bytes27,
    bytes28,
    bytes29,
    bytes30,
    bytes31,
    bytes32,
    bytes,
    string,
    str,
    address,
    bool,
    HexAddress,
    ChecksumAddress,
    HexStr,
]

# in solidity, these can not be assigned to yet but can be declared
ufixed128x18 = NewType("ufixed128x18", float)
fixed128x18 = NewType("fixed128x18", float)



from inspect import isclass
from types import GenericAlias
from typing import Generic, TypeVar, cast, get_args, get_origin

from eth_abi import decode, encode
from eth_hash.auto import keccak as keccak_256
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

from .._request import Request
from ..types import BASIC_TYPES, Name, NoArgs, Struct
from ..utils import is_annotation, transform_primitive

T = TypeVar(
    "T",
    bound=tuple
    | BaseModel
    | BASIC_TYPES
    | list[BASIC_TYPES]
    | tuple[BASIC_TYPES | Struct, ...]
    | HexAddress,
)
U = TypeVar("U")


class FuncSignature(Request, Generic[T, U]):
    alias: str | None = None
    name: str

    def get_identifier(self):
        """This works most of the time"""
        signature = f'{self.name}({",".join(self.get_inputs())})'
        return f"0x{keccak_256(signature.encode('utf-8')).hex()[:8]}"

    def get_inputs(self):
        from ..types import Struct

        inputs, _ = self.__pydantic_generic_metadata__["args"]
        if inputs is NoArgs:
            return []
        if (
            type(inputs) is not GenericAlias
            and isclass(inputs)
            and issubclass(inputs, BaseModel)
        ):
            converted_inputs = transform_primitive(inputs)
            if issubclass(inputs, Struct):
                return [converted_inputs]
        else:
            if get_origin(inputs) == list:
                return [transform_primitive(inputs)]
            elif get_origin(inputs) == tuple:
                return tuple(transform_primitive(input) for input in get_args(inputs))
            else:
                converted_inputs = transform_primitive(inputs)
        if not isinstance(converted_inputs, list):
            return [converted_inputs]
        return converted_inputs

    @property
    def _output(self):
        output_type = self.__pydantic_generic_metadata__["args"][1]
        if is_annotation(output_type):
            return get_args(output_type)[0]
        return output_type

    @property
    def _inputs(self):
        input_type = self.__pydantic_generic_metadata__["args"][0]
        if is_annotation(input_type):
            return get_args(input_type)[0]
        return input_type

    def get_output(self):
        outputs = self._output
        if outputs is type(None):
            return None

        if is_annotation(outputs):
            outputs = get_args(outputs)[0]

        if (
            isclass(outputs)
            and not isinstance(outputs, GenericAlias)
            and issubclass(outputs, BaseModel)
        ):
            converted_outputs = transform_primitive(outputs)
        else:
            converted_outputs = transform_primitive(outputs)
        return converted_outputs

    def get_output_name(self):
        return [self._get_name(output) for output in get_args(self._output)]

    @classmethod
    def _encode(cls, value):
        if isinstance(value, tuple):
            return tuple(cls._encode(val) for val in value)
        elif isinstance(value, list):
            return [cls._encode(val) for val in value]
        elif isinstance(value, BaseModel):
            return tuple(cls._encode(val) for val in value.model_dump().values())
        return value

    def encode_call(self, *, inputs: T) -> HexStr:
        from ..types import Struct

        identifier = self.get_identifier()
        # TODO: this is hard
        if inputs == ():
            return identifier

        if isinstance(inputs, BaseModel):
            if isinstance(inputs, Struct):
                input_data = inputs.to_bytes().hex()
            else:
                input_data = encode(
                    self.get_inputs(),
                    list(inputs.model_dump().values()),
                ).hex()
        elif isinstance(inputs, tuple):
            input_data = encode(
                self.get_inputs(), [self._encode(val) for val in inputs]
            ).hex()
        elif isinstance(inputs, list):
            input_data = encode(
                self.get_inputs(), [[self._encode(val) for val in inputs]]
            ).hex()
        else:
            input_data = encode(self.get_inputs(), [inputs]).hex()
        return HexStr(f"{identifier}{input_data}")

    def decode_result(self, result: HexStr) -> U:
        if isclass(self._output) and issubclass(self._output, Struct):
            return self._output.from_bytes(result)

        output = self.get_output()
        if output is None:
            # return None if the expected return type is None
            return output  # type: ignore

        if not isinstance(output, list):
            output = [output]
            decoded_output = decode(output, bytes.fromhex(result.removeprefix("0x")))[0]

            if get_origin(self._output) == list:
                output_list_type = get_args(self._output)[0]
                if isclass(output_list_type) and issubclass(output_list_type, Struct):
                    decoded_arr: U = cast(
                        U,
                        [output_list_type.from_tuple(item) for item in decoded_output],
                    )
                    return decoded_arr
        else:
            if isclass(self._output) and issubclass(self._output, Struct):
                return self._output.from_bytes(bytes.fromhex(result.removeprefix("0x")))
            else:
                decoded_output = decode(
                    output, bytes.fromhex(result.removeprefix("0x"))
                )

        # NOTE: https://github.com/pydantic/pydantic/discussions/5970
        # TODO: this is discussed to see if its a bug or not.  Annotations are a class but can't be checked as a subclass
        if (
            isclass(self._output)
            and not isinstance(self._output, GenericAlias)
            and issubclass(self._output, BaseModel)
        ):
            decoded = decoded_output
            output = self._output
            response = {}
            for (name, field), value in zip(output.model_fields.items(), decoded):
                response[name] = Struct.cast(field.annotation, value)
            return output(**response)
        return decoded_output

    @staticmethod
    def _get_name(type):
        if is_annotation(type):
            type, *annotations = get_args(type)
            for annotation in annotations:
                if isinstance(annotation, Name):
                    return annotation.value
        return ""



import json

import click
from eth_typing import HexAddress

from .block_explorer import get_abi
from .codegen import codegen as codegen_cmd


@click.group()
def cli():
    pass


@click.group()
def codegen():
    pass


@click.command()
@click.argument("input_file")
@click.option("--output", "-o", default="abi.py")
@click.option("--contract-name", "-c", default="AnonContract")
def load(input_file, output, contract_name: str):
    with open(input_file, "r") as f:
        raw_abi = f.read()

    abi = json.loads(raw_abi)
    if isinstance(abi, dict):
        try:
            abi = abi["abi"]
        except KeyError:
            raise ValueError("Invalid Format")
    with open(output, "w") as f:
        f.write(codegen_cmd(abi, contract_name))


@click.command()
@click.option("--network", "-n", default="ethereum")
@click.option("--address", "-a", required=True)
@click.option("--output", "-o", default="abi.py")
@click.option("--api_key", "-a", required=True)
@click.option("--contract-name", "-c", default="AnonContract")
def explorer(
    network: str, address: HexAddress, output: str, api_key, contract_name: str
):
    if network.lower() != "ethereum":
        click.echo("Network not yet supported.  Coming soon!")
        return
    abi = get_abi(address, api_key)
    with open(output, "w") as f:
        f.write(codegen_cmd(abi, contract_name))


cli.add_command(codegen)
codegen.add_command(load)
codegen.add_command(explorer)

if __name__ == "__main__":
    cli()



import os

import pytest
from eth_rpc import Log, set_alchemy_key
from eth_rpc.networks import Arbitrum, Ethereum


@pytest.mark.unit
def test_log_network():
    assert Log[Ethereum]._network == Ethereum
    assert Log[Arbitrum]._network == Arbitrum


@pytest.mark.asyncio(scope="session")
async def test_log():
    set_alchemy_key(os.environ["ALCHEMY_KEY"])
    eth_logs = await Log[Ethereum].load_by_number(10_000_000, 10_000_000)
    arb_logs = await Log[Arbitrum].load_by_number(10_000_000, 10_000_000)

    assert len(eth_logs) == 135
    assert len(arb_logs) == 4



from collections.abc import Awaitable
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Annotated,
    ClassVar,
    Literal,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from eth_hash.auto import keccak as keccak_256
from eth_rpc.types import (
    BLOCK_STRINGS,
    ContractMethod,
    GetCodeArgs,
    GetStorageArgs,
    HexInteger,
    MaybeAwaitable,
)
from eth_typing import HexAddress, HexStr, Primitives
from pydantic import BaseModel, Field

from .._request import Request
from ..utils import run, to_hex_str
from .function import ContractFunc
from .interface import ContractT

if TYPE_CHECKING:
    from .function import FuncSignature


T = TypeVar(
    "T",
    bound=tuple
    | BaseModel
    | Primitives
    | list[Primitives]
    | tuple[Primitives, ...]
    | HexAddress,
)
U = TypeVar("U")


class Contract(ContractT, Request):
    _func_sigs: ClassVar[dict[str, ContractMethod]]
    functions: list[ContractFunc] = Field(default_factory=list)

    @property
    def sync(self) -> "ContractSync":
        obj = deepcopy(self)
        obj.__class__ = ContractSync
        return obj  # type: ignore

    @classmethod
    def __get_parent_func_sigs(cls) -> dict[str, ContractMethod]:
        func_sigs: dict[str, ContractMethod] = {}
        for klass in cls.__mro__:
            if issubclass(klass, Contract):
                try:
                    func_sigs |= klass._func_sigs
                except AttributeError:
                    pass
        return func_sigs

    def __init_subclass__(cls, **kwargs) -> None:
        # TODO: I hate this slightly less
        del_keys = []

        cls_annotations = get_type_hints(cls, include_extras=True)
        annotations = {
            key: cls_annotations[key]
            for key in (cls_annotations.keys() - get_type_hints(Contract).keys())
        }
        cls._func_sigs = cls.__get_parent_func_sigs()

        for key, annotation in annotations.items():
            if key.startswith("_"):  # exclude protected/private attributes
                continue
            if get_origin(annotation) is Annotated:
                annotation_args = get_args(annotation)
                for annotation in annotation_args:
                    if (
                        annotation is ContractMethod
                        or get_origin(annotation) is ContractFunc
                    ):
                        cls._func_sigs[key] = cls_annotations[key]
                        del_keys.append(key)
                        break
            elif get_origin(annotation) is ContractFunc:
                cls._func_sigs[key] = cls_annotations[key]
                del_keys.append(key)
        for key in del_keys:
            del cls.__annotations__[key]

    def add_func(self, func: "FuncSignature"):
        if func not in self.functions:
            self.functions.append(ContractFunc(func=func, contract=self))

    async def _get_storage_at(
        self, slot: int | HexStr, block_number="latest", sync: bool = False
    ):
        if sync:
            return self.rpc().get_storage_at.sync(
                GetStorageArgs(
                    storage_address=self.address,
                    slot_position=to_hex_str(slot),
                    block_number=block_number,
                )
            )
        return await self.rpc().get_storage_at(
            GetStorageArgs(
                storage_address=self.address,
                slot_position=to_hex_str(slot),
                block_number=block_number,
            )
        )

    @overload
    def get_storage_at(
        self, *, sync: Literal[True], slot: int | HexStr, block_number="latest"
    ) -> HexStr: ...

    @overload
    def get_storage_at(
        self, *, slot: int | HexStr, block_number="latest"
    ) -> Awaitable[HexStr]: ...

    def get_storage_at(
        self, *, slot: int | HexStr, block_number="latest", sync: bool = False
    ) -> MaybeAwaitable[HexStr]:
        return run(
            self._get_storage_at,
            slot=slot,
            block_number=block_number,
            sync=sync,
        )

    async def _get_code(
        self,
        block_number: int | BLOCK_STRINGS | None = None,
        block_hash: HexStr | None = None,
        sync: bool = False,
    ) -> HexStr:
        if block_hash:
            if sync:
                return self.rpc().get_code.sync(
                    GetCodeArgs(
                        address=self.address,
                        block_hash=block_hash,
                    )
                )
            return await self.rpc().get_code(
                GetCodeArgs(
                    address=self.address,
                    block_hash=block_hash,
                )
            )

        if block_number is None:
            block_number = "latest"
        if sync:
            return self.rpc().get_code.sync(
                GetCodeArgs(
                    address=self.address,
                    block_number=(
                        HexInteger(block_number)
                        if isinstance(block_number, int)
                        else block_number
                    ),
                )
            )
        return await self.rpc().get_code(
            GetCodeArgs(
                address=self.address,
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
            )
        )

    @overload
    def get_code(
        self,
        *,
        sync: Literal[True],
        block_number: int | BLOCK_STRINGS | None = None,
        block_hash: HexStr | None = None,
    ) -> HexStr: ...

    @overload
    def get_code(
        self,
        *,
        block_number: int | BLOCK_STRINGS | None = None,
        block_hash: HexStr | None = None,
    ) -> Awaitable[HexStr]: ...

    def get_code(
        self,
        *,
        block_number: int | BLOCK_STRINGS | None = None,
        block_hash: HexStr | None = None,
        sync: bool = False,
    ) -> MaybeAwaitable[HexStr]:
        return run(
            self._get_code,
            block_number=block_number,
            block_hash=block_hash,
            sync=sync,
        )

    def create2(self, salt: bytes, keccak_init_code: bytes) -> HexAddress:
        """
        EIP-104
        https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1014.md
        """
        pre = "0xff"
        b_pre = bytes.fromhex(pre[2:])
        b_address = bytes.fromhex(self.address[2:])

        b_result = keccak_256(b_pre + b_address + salt + keccak_init_code)
        result_address = "0x" + b_result[12:].hex()

        return HexAddress(HexStr(result_address))


class ContractSync(Contract):
    SYNC: ClassVar[Literal[True]] = True

    def get_storage_at(  # type: ignore
        self,
        *,
        slot: int | HexStr,
        block_number="latest",
    ) -> HexStr:
        return super().get_storage_at(
            slot=slot,
            block_number=block_number,
            sync=self.SYNC,
        )

    def get_code(  # type: ignore
        self,
        *,
        block_number: int | BLOCK_STRINGS | None = None,
        block_hash: HexStr | None = None,
    ) -> HexStr:
        return super().get_code(
            block_number=block_number, block_hash=block_hash, sync=self.SYNC
        )



from typing import Any, Optional

from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel

from ..basic import HexInteger


class AlchemyParams(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    block_number: Optional[HexInteger]
    block_hash: Optional[HexStr]


class AlchemyBlockReceipt(BaseModel):
    params: AlchemyParams

    def model_dump(self, exclude_none=True, by_alias=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )


class AlchemyTokenBalance(BaseModel):
    contract_address: HexAddress = Field(alias="contractAddress")
    token_balance: HexStr | None = Field(alias="tokenBalance", default=None)
    error: str | None = Field(default=None)


class AlchemyTokenBalances(BaseModel):
    address: HexAddress
    token_balances: list[AlchemyTokenBalance] = Field(alias="tokenBalances")



from typing import ClassVar, Optional

from eth_typing import HexAddress, HexStr
from pydantic import AnyHttpUrl, BaseModel, Field
from pydantic.networks import AnyWebsocketUrl, Url


class BlockExplorer(BaseModel):
    name: str
    url: str
    api_url: str
    api_key: str | None = None


class RpcUrl(BaseModel):
    http: AnyHttpUrl
    wss: Optional[AnyWebsocketUrl] = Field(default=None)


class Rpcs(BaseModel):
    default: RpcUrl
    backups: list[RpcUrl] = []


class Network(BaseModel):
    http: ClassVar[str]
    wss: ClassVar[str]

    chain_id: ClassVar[int]
    name: ClassVar[str]
    native_currency: ClassVar[str]
    rpc: ClassVar[Rpcs]
    block_explorer: ClassVar[BlockExplorer]
    alchemy_str: ClassVar[str | None] = None
    multicall3: ClassVar[HexAddress | None] = HexAddress(
        HexStr("0xca11bde05977b3631167028862be2a173976ca11")
    )
    ens_registry: ClassVar[HexAddress | None] = HexAddress(
        HexStr("0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e")
    )
    ens_universal_resolver: ClassVar[HexAddress | None] = HexAddress(
        HexStr("0xce01f8eee7E479C928F8919abD53E553a36CeF67")
    )
    apprx_block_time: ClassVar[float] = 12.0

    @classmethod
    def set(
        cls,
        http: str | None = None,
        wss: str | None = None,
        api_key: str | None = None,
    ):
        if http:
            cls.rpc.default.http = Url(http)
            cls.http = str(cls.rpc.default.http)
        if wss:
            cls.rpc.default.wss = Url(wss)
            cls.wss = str(cls.rpc.default.wss)
        if api_key:
            cls.block_explorer.api_key = api_key
        return cls

    def __init_subclass__(cls, **kwargs):
        if cls.rpc.default.http:
            cls.http = cls.rpc.default.http
        if cls.rpc.default.wss:
            cls.wss = cls.rpc.default.wss

    def __hash__(self) -> int:
        return self.chain_id

    def __str__(self):
        return f"<Network: {self.name}>"

    __repr__ = __str__



from typing import ClassVar

from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url


class Linea(Network):
    chain_id: ClassVar[int] = 59144
    name: ClassVar[str] = "Linea"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://rpc.linea.build"),
            wss=Url("wss://rpc.linea.build"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Lineascan",
        url="https://lineascan.build",
        api_url="https://api.lineascan.build/api",
    )
    alchemy_str: ClassVar[str | None] = "linea-mainnet"
    apprx_block_time: ClassVar[float] = 2.0



import re
from typing import Any

from eth_rpc.types import NoArgs, primitives


def to_snake_case(name):
    if name.isupper():
        return name
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


def _convert_type(type_: str) -> Any:
    if type_ in ["bytes", "bool"]:
        return type_
    if type_.endswith("[]"):
        return list[_convert_type(type_[:-2])]  # type: ignore
    for _type in dir(primitives):
        if _type == type_:
            return getattr(primitives, type_)
    if type_.startswith("enum "):
        return primitives.uint8
    raise ValueError(f"Invalid Type {type_}")


def object_to_type(obj):
    if "components" in obj:
        components = [object_to_type(t) for t in obj["components"]]
        return tuple[*components]
    try:
        return _convert_type(obj["internalType"])
    except ValueError:
        return _convert_type(obj["type"])


def convert_types(types_):
    lst = []
    models = []
    for type_ in types_:
        if "components" in type_:
            py_model_name = type_["internalType"].split(".")[-1].removeprefix("struct ")
            field_name = py_model_name
            while field_name.endswith("[]"):
                field_name = f"list[{py_model_name[:-2]}]"
            lst.append(field_name)
            models.append((py_model_name, type_["components"]))
        else:
            lst.append(object_to_type(type_))
    if len(lst) == 0:
        return lst, models
    if len(lst) == 1:
        return lst[0], models
    return (tuple[*lst], models)


def codegen(abi: list[dict[str, Any]], contract_name: str) -> str:  # noqa: C901
    """
    Convert an ABI to the string implementation of a ProtocolBase.

    Make sure to convert it to snakecase and use the Annotated[type, Name("name")] syntax for the contract func

    For Example:
    [{"inputs":[],"name":"WETH","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"}]

    would convert to:

    from eth_rpc import ProtocolBase, ContractFunc
    from eth_rpc.types import primitives, NoArgs, Struct

    class WETH(ProtocolBase):
        WETH: ContractFunc[NoArgs, primitives.address]
    """
    # Begin building the class string
    imports = [
        "from typing import Annotated\n",
        "from eth_rpc import ProtocolBase, ContractFunc",
        "from eth_rpc.types import primitives, Name, NoArgs, Struct",
        "\n",
    ]
    lines = [
        f"\nclass {contract_name}(ProtocolBase):",
    ]

    # Iterate over each function in the ABI
    model_dict = {}
    for func in abi:
        if func["type"] != "function":
            continue  # Skip non-function types

        func_name = func.get("name", "unnamed_function")
        inputs = func.get("inputs", [])
        outputs = func.get("outputs", [])

        input_type, _models = convert_types(inputs)
        for model in _models:
            if model[0] not in model_dict:
                model_dict[model[0]] = model[1]
            elif model_dict[model[0]] == model[1]:
                continue
            else:
                print("Warning: Duplicate model name with different fields")
                model_dict[model[0] + "_extra"] = model[1]

        output_type, __models = convert_types(outputs)

        for model in __models:
            model_name = model[0].replace("[]", "")
            if model_name not in model_dict:
                model_dict[model_name] = model[1]
            elif model_dict[model_name] == model[1]:
                continue
            else:
                print("Warning: Duplicate model name with different fields")
                model_dict[model_name + "_extra"] = model[1]

        has_name_annotation: bool = False
        alias: str
        if func_name != (
            py_name := to_snake_case(
                func_name.replace("WETH", "Weth").replace("ETH", "Eth")
            )
        ):
            has_name_annotation = True
            alias = py_name
        else:
            alias = func_name

        # Append the function definition to the class
        if input_type == NoArgs or input_type == []:
            input_type = "NoArgs"
        if output_type == []:
            output_type = "None"

        input_type = str(input_type).replace("'", "")
        output_type = str(output_type).replace("'", "")
        if not has_name_annotation:
            lines.append(
                f"\t{func_name}: ContractFunc[\n\t\t{input_type},\n\t\t{output_type}\n\t]\n"
            )
        else:
            lines.append(
                f'\t{alias}: Annotated[\n\t\tContractFunc[\n\t\t\t{input_type},\n\t\t\t{output_type}\n\t\t],\n\t\tName("{func_name}"),\n\t]\n'
            )

    # Combine all lines into a single string

    models = list(model_dict.items())
    for name, fields in models:
        model_str = f"""
class {name}(Struct):
"""
        embedded_types = []
        for field in fields:
            try:
                type_ = _convert_type(field["internalType"])
            except ValueError:
                type_ = object_to_type(field)
                embedded_types.append(type_)
            model_str += f"\t{field['name']}: {type_}\n"
        lines = [model_str] + lines

    class_str = "\n".join(imports) + "\n".join(lines) + "\n"
    return class_str.replace("eth_rpc.types.", "").replace("\t", "    ").strip() + "\n"



import pytest
from eth_rpc import Transaction
from eth_rpc.networks import Arbitrum, Ethereum
from eth_typing import HexStr


@pytest.mark.unit
def test_transaction_network() -> None:
    # Currently, block is a singleton
    tx: type[Transaction[Arbitrum]] = Transaction[Arbitrum]
    assert tx._network == Arbitrum
    assert Transaction[Ethereum]._network == Ethereum


@pytest.mark.asyncio(scope="session")
async def test_transaction_load() -> None:
    # Currently, block is a singleton
    tx = await Transaction[Ethereum].get_by_hash(
        HexStr("0x38cab69c76d8d7069f8b90e170a853740f6370f975de9f4d90f50ed6adc2e6ee")
    )
    assert tx
    assert tx.block_number == 20590874
    assert tx.gas_price == 2800000000

    tx2 = await Transaction[Arbitrum].get_by_hash(
        HexStr("0x71183e2c32180bceca2696496e98a6d2805b03676df2cf0ab5189f9169514b16")
    )
    assert tx2
    assert tx2
    assert tx2.block_number == 242473407
    assert tx2.gas_price == 13000000

    # make sure they keep their networks
    assert tx._network == Ethereum
    assert tx2._network == Arbitrum



from eth_rpc.types import HexInteger
from eth_rpc.utils import RPCModel
from eth_typing import HexAddress, HexStr


class AccessList(RPCModel):
    address: HexAddress
    storage_keys: list[HexStr]


class AccessListResponse(RPCModel):
    access_list: list[AccessList]
    gas_used: HexInteger



from typing import Annotated

import pytest
from eth_abi import encode
from eth_rpc import FuncSignature
from eth_rpc.types import Name, NoArgs, Struct, primitives
from eth_rpc.utils.types import transform_primitive
from eth_typing import HexStr
from pydantic import BaseModel


class Inputs(BaseModel):
    x: primitives.uint256
    y: primitives.uint256
    z: list[primitives.address]
    a: primitives.address
    b: primitives.uint256
    c: list[primitives.address]
    d: primitives.uint256


class Outputs(BaseModel):
    x: Annotated[
        list[primitives.bytes32],
        Name("XX"),
    ]
    y: str


class DisputeNested(Struct):
    a: list[primitives.address]
    b: list[primitives.uint256]
    c: primitives.uint256
    d: primitives.uint16


class DisputeInput(BaseModel):
    first: DisputeNested
    second: primitives.address
    third: primitives.address


@pytest.mark.unit
def test_func_signature():
    func = FuncSignature[Inputs, Annotated[Outputs, Name("OUTPUT")]](
        name="swapExactTokensForETHWithPortfolios", alias="test_func"
    )
    # https://www.4byte.directory/signatures/?bytes4_signature=0x8b41f64c
    assert func.get_identifier() == "0x8b41f64c"
    assert func.get_inputs() == [
        "uint256",
        "uint256",
        "address[]",
        "address",
        "uint256",
        "address[]",
        "uint256",
    ]
    assert func.get_output() == ["bytes32[]", "string"]
    assert func._get_name(Outputs.__annotations__["x"]) == "XX"

    func = FuncSignature[
        tuple[int, list[primitives.address], list[bytes]],
        bool,
    ](name="makeProxyArbitraryTransactions")
    assert func.get_inputs() == ("uint256", "address[]", "bytes[]")
    assert func.get_identifier() == "0xbe430874"

    func = FuncSignature[
        tuple[list[primitives.address], list[primitives.uint96], list[bool]],
        bool,
    ](name="onBridgeOperatorsAdded")
    assert func.get_identifier() == "0x8f851d8a"

    func3 = FuncSignature[
        DisputeInput,
        bool,
    ](name="distribute")
    assert func3.get_identifier() == "0x2d3f5537"


@pytest.mark.unit
def test_map_to_str():
    assert transform_primitive(primitives.address) == "address"
    assert transform_primitive(str) == "string"
    assert transform_primitive(list[str]) == "string[]"
    assert transform_primitive(list[primitives.bytes32]) == "bytes32[]"


@pytest.mark.unit
def test_func_signature_types():
    func = FuncSignature[
        tuple[list[int], bool],
        list[str],
    ](name="func")
    assert func.get_inputs() == ("uint256[]", "bool")
    assert func.get_output() == "string[]"


@pytest.mark.unit
def test_func_encode() -> None:
    class Data(Struct):
        name: str
        valid: bool

    func = FuncSignature[
        tuple[primitives.bytes32, list[Data]],
        None,
    ](name="func")
    assert func.get_inputs() == ("bytes32", "(string,bool)[]")
    func.encode_call(
        inputs=(
            primitives.bytes32(b"123"),
            [Data(name="test", valid=True), Data(name="test2", valid=False)],
        )
    )


@pytest.mark.unit
def test_func_decode_struct_list() -> None:
    class MyNestedStruct(Struct):
        a: primitives.uint256
        b: primitives.uint256

    class MyStruct(Struct):
        x: primitives.uint256
        y: bool
        z: list[MyNestedStruct]

    func = FuncSignature[
        NoArgs,
        list[MyStruct],
    ](name="func")

    my_struct = MyStruct(x=1, y=True, z=[MyNestedStruct(a=1, b=2)])
    encoded_bytes = encode(
        ["(uint256,bool,(uint256,uint256)[])[]"], [[(1, True, [(1, 2)])]]
    )

    result = func.decode_result(HexStr(encoded_bytes.hex()))
    assert result[0] == my_struct



from ._transport import (
    configure_rpc_from_env,
    get_current_network,
    get_selected_wallet,
    set_alchemy_key,
    set_default_network,
    set_rpc_timeout,
    set_rpc_url,
    set_selected_wallet,
    set_transport,
)
from .account import Account
from .block import Block
from .codegen import codegen
from .contract import Contract, ContractFunc, EthResponse, FuncSignature, ProtocolBase
from .event import Event
from .log import Log
from .models import EventData
from .rpc import Middleware, add_middleware
from .subscriber import EventSubscriber
from .transaction import PreparedTransaction, Transaction, TransactionReceipt
from .types import Network
from .wallet import PrivateKeyWallet

# we need to rebuild block because we use a ForwardRef for Transactions
Block.model_rebuild()


__all__ = [
    "Account",
    "Block",
    "Contract",
    "ContractFunc",
    "EthResponse",
    "Event",
    "EventData",
    "EventSubscriber",
    "FuncSignature",
    "Log",
    "Middleware",
    "Network",
    "PreparedTransaction",
    "PrivateKeyWallet",
    "ProtocolBase",
    "Transaction",
    "TransactionReceipt",
    "add_middleware",
    "codegen",
    "configure_rpc_from_env",
    "get_current_network",
    "get_selected_wallet",
    "set_alchemy_key",
    "set_default_network",
    "set_selected_wallet",
    "set_transport",
    "set_rpc_timeout",
    "set_rpc_url",
]



from typing import ClassVar

from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url


class Sapphire(Network):
    chain_id: ClassVar[int] = 23294
    name: ClassVar[str] = "Sapphire"
    native_currency: ClassVar[str] = "ROSE"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://sapphire.oasis.io	"),
            wss=Url("wss://sapphire.oasis.io/ws	"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Sapphire Testnet",
        url="https://explorer.oasis.io/mainnet/sapphire",
        api_url="https://nexus.oasis.io/v1/sapphire",
    )
    alchemy_str: ClassVar[str | None] = None
    apprx_block_time: ClassVar[float] = 6


class SapphireTestnet(Network):
    chain_id: ClassVar[int] = 23295
    name: ClassVar[str] = "Sapphire Testnet"
    native_currency: ClassVar[str] = "ROSE"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://testnet.sapphire.oasis.io"),
            wss=Url("wss://testnet.sapphire.oasis.io/ws"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Sapphire Testnet",
        url="https://explorer.oasis.io/testnet/sapphire",
        api_url="https://testnet.nexus.oasis.io/v1/sapphire",
    )
    alchemy_str: ClassVar[str | None] = None
    apprx_block_time: ClassVar[float] = 6



class LogResponseExceededError(Exception):
    message: str
    recommended_start: int
    recommended_end: int

    def __init__(self, message, recommended_start: int, recommended_end: int):
        self.message = message
        super().__init__(message)
        self.recommended_start = recommended_start
        self.recommended_end = recommended_end


class LogDecodeError(ValueError): ...


class RateLimitingError(ValueError): ...


class RPCDecodeError(ValueError): ...


class UnsupportedChainIDException(ValueError): ...



from typing import Awaitable, Callable, ClassVar, Generic, ParamSpec

from pydantic import BaseModel

from .base import Params, Response, RPCMethodBase

P = ParamSpec("P")


class Middleware(BaseModel):
    def update(self, method: "RPCMethod", make_request, params=None): ...


class RPCMethod(RPCMethodBase[Params, Response], Generic[Params, Response]):
    middlewares: ClassVar[list[Callable]] = []

    @classmethod
    def add_middleware(cls, middleware: Callable):
        cls.middlewares.append(middleware)

    def __call__(self, *params: Params) -> Awaitable[Response]:
        async def run():
            make_request = self.call_async
            for middleware in self.middlewares:
                make_request = middleware(self, make_request, is_async=True)
            return await make_request(*params)

        return run()

    @property
    def sync(self) -> Callable[..., Response]:
        make_request = self.call_sync

        for middleware in self.middlewares:
            make_request = middleware(self, make_request)

        return make_request


def add_middleware(middleware: list[Callable] | Callable):
    if not isinstance(middleware, list):
        middleware = [middleware]

    for elem in middleware:
        RPCMethod.add_middleware(elem)



from enum import Enum
from inspect import isclass
from types import GenericAlias
from typing import Annotated, get_args, get_origin

from eth_abi import encode
from eth_hash.auto import keccak
from eth_rpc.types import ALL_PRIMITIVES, NoArgs, Struct
from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import BaseModel


def is_annotation(type_):
    return get_origin(type_) is Annotated


def transform_primitive(type_, with_name: bool = False, name: str = ""):
    if type_ in [(), NoArgs]:
        return None
    type_ = _transform_primitive(type_)
    if with_name:
        return f"{type_} {name}"
    return type_


def _transform_primitive(type_):
    mapping = {
        str: "string",
        int: "uint256",  # defaults to uint
        HexAddress: "address",
        ChecksumAddress: "address",
        HexStr: "bytes",
    }
    if type_ in mapping:
        return mapping[type_]
    if type_.__name__ in ALL_PRIMITIVES:
        return type_.__name__

    if isclass(type_) and issubclass(type_, Enum):
        return "uint8"

    if is_annotation(type_):
        return transform_primitive(get_args(type_)[0])
    if isinstance(type_, GenericAlias):
        if get_origin(type_) == list:
            (arg,) = get_args(type_)
            return f"{transform_primitive(arg)}[]"
        elif get_origin(type_) == tuple:
            args = get_args(type_)
            tuple_args = ",".join(transform_primitive(a) for a in args)
            return f"({tuple_args})"
        else:
            raise ValueError(f"Unknown GenericAlias: {type_}")
    # elif isclass(type_) and issubclass(type_, Struct):
    elif isclass(type_) and issubclass(type_, Struct):
        # If this is a Struct, return a tuple
        args = [arg.annotation for arg in type_.model_fields.values()]
        return transform_primitive(tuple[*args])
    elif isclass(type_) and issubclass(type_, BaseModel):
        # If this is a BaseModel and NOT a Struct, return a list
        args = [arg.annotation for arg in type_.model_fields.values()]
        return [transform_primitive(arg) for arg in args]
    return transform_primitive(type_)


def to_hex_str(number: int | str) -> HexStr:
    if isinstance(number, int):
        return HexStr(hex(number))
    return HexStr(number)


def to_bytes32(input_: int | HexStr):
    if isinstance(input_, str):
        return encode(["bytes32"], [bytes.fromhex(input_.lstrip("0x"))])
    return encode(["bytes32"], [input_.to_bytes()])


def to_topic(event_sig: str) -> bytes:
    return keccak(event_sig.replace(" ", "").encode("utf-8"))



from asyncio import iscoroutine
from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any, TypeVar, cast

if TYPE_CHECKING:
    from eth_rpc.types import MaybeAwaitable

T = TypeVar("T")


async def handle_maybe_awaitable(t: "MaybeAwaitable[T]") -> T:
    if iscoroutine(t):
        return await t
    else:
        return cast(T, t)


def run(
    corofunc: Callable[..., Coroutine[Any, Any, T]], *args, sync: bool = False, **kwargs
) -> T | Coroutine[Any, Any, T]:
    if sync:
        try:
            next(corofunc(*args, sync=sync, **kwargs).__await__())
        except StopIteration as ex:
            return ex.value
        raise RuntimeError("Not a sync execution")
    return corofunc(*args, sync=sync, **kwargs)


class DualAsync:
    @staticmethod
    def dual(method, sync: bool = False):
        def _method(self: "DualAsync", *args, **kwargs):
            return method(self, *args, is_async=not sync, **kwargs)

        return _method



from eth_hash.auto import keccak
from eth_typing import ChecksumAddress, HexAddress, HexStr


def address_to_topic(address: str) -> HexStr:
    return HexStr("0x{:064x}".format(int(address, 16)))


def to_checksum(address: str) -> ChecksumAddress:
    norm_address = address.lower()
    address_hash = "0x" + keccak(norm_address[2:].encode("utf-8")).hex()
    return ChecksumAddress(
        HexAddress(
            HexStr(
                "0x"
                + "".join(
                    (
                        norm_address[i].upper()
                        if int(address_hash[i], 16) > 7
                        else norm_address[i]
                    )
                    for i in range(2, 42)
                )
            )
        )
    )



import math
from typing import Annotated, Any, Literal, get_args

from pydantic import (
    BeforeValidator,
    GetCoreSchemaHandler,
    SerializerFunctionWrapHandler,
    ValidationInfo,
)
from pydantic.functional_serializers import WrapSerializer
from pydantic_core import core_schema

from . import primitives

ALL_PRIMITIVES = [x for x in dir(primitives) if x[0].islower() and x[0] != "_"]
BLOCK_STRINGS = Literal["latest", "earliest", "pending", "safe", "finalized"]


class HexInt(int):
    """
    This enables an integer to load from a hex str to an int, or to be loaded as an int directly
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            function=cls.validate,
            schema=handler(int | str),
            field_name=handler.field_name,
        )

    @classmethod
    def validate(cls, v, r: ValidationInfo):
        if isinstance(v, str):
            return int(v, 16)
        return int(v)


# wraps an int to a hex, or preserves the block_string
def hex_int_wrap(v: Any, nxt: SerializerFunctionWrapHandler) -> str:
    if v in get_args(BLOCK_STRINGS):
        return v
    return hex(v)


def number_to_bytes(number):
    nibble_count = int(math.log(number, 256)) + 1
    hex_string = "{:0{}x}".format(number, nibble_count * 2)
    return bytes.fromhex(hex_string)[:32]


def hex_str_to_bytes(v, info):
    if isinstance(v, bytes):
        return v
    elif v == "":
        return b""
    return number_to_bytes(int(v, 16))


# A Custom Integer type that loads from hex string and serializes to hex string
# TODO: make this a BlockNumber type and support block strings too
HexInteger = Annotated[HexInt, WrapSerializer(hex_int_wrap)]

# references to a block, either the rpc terms or a HexInteger
BlockReference = HexInteger | BLOCK_STRINGS

Bytes32Hex = Annotated[
    primitives.bytes32,
    BeforeValidator(hex_str_to_bytes),
    WrapSerializer(lambda x, y, z: x.hex()),
]



import json

import httpx
from eth_typing import HexAddress


def get_abi(
    address: HexAddress, api_key: str, api_url: str = "https://api.etherscan.io/api"
):
    response = httpx.get(
        f"{api_url}?module=contract&action=getabi&address={address}&apikey={api_key}"
    )
    return json.loads(response.json()["result"])



import asyncio
import json
import logging
from collections.abc import AsyncIterator
from typing import Generic, Optional, cast

from eth_rpc.models import AccessList, PendingTransaction
from eth_rpc.models import Transaction as TransactionModel
from eth_rpc.models import TransactionReceipt as TransactionReceiptModel
from eth_rpc.types.args import (
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    RawTransaction,
    TransactionRequest,
)
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict, model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import TypeVar
from websockets.exceptions import ConnectionClosedError
from websockets.legacy.client import WebSocketClientProtocol, connect

from ._request import Request
from ._transport import _force_get_global_rpc
from .types import (
    BLOCK_STRINGS,
    AlchemyBlockReceipt,
    AlchemyParams,
    HexInteger,
    JsonPendingWssResponse,
    RPCResponseModel,
    SubscriptionResponse,
)

T = TypeVar("T")
Network = TypeVar("Network", default=None)

logger = logging.getLogger(__name__)


class PreparedTransaction(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    type: int = 2
    gas: int
    gas_price: int | None = None
    max_fee_per_gas: int | None = None
    max_priority_fee_per_gas: int | None = None
    data: HexStr
    nonce: int
    to: HexAddress
    value: int
    access_list: Optional[list[AccessList]] = None
    chain_id: int

    def model_dump(self, *args, exclude_none=True, by_alias=True, **kwargs):
        return super().model_dump(
            *args, exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )

    @model_validator(mode="after")
    def validate_xor(self):
        if self.max_fee_per_gas is not None:
            self.type = 2
        else:
            self.type = 1
        if self.gas_price is None and self.max_fee_per_gas is None:
            raise ValueError("Must set gas_price or max_fee_per_gas")
        if self.max_fee_per_gas is not None and self.max_priority_fee_per_gas is None:
            self.max_priority_fee_per_gas = 0
        return self


class TransactionReceipt(Request, TransactionReceiptModel, Generic[Network]):
    @classmethod
    def get_by_hash(
        self, tx_hash: HexStr
    ) -> RPCResponseModel[TransactionRequest, Optional[TransactionReceiptModel]]:
        return RPCResponseModel(
            self.rpc().get_tx_receipt,
            TransactionRequest(
                tx_hash=tx_hash,
            ),
        )

    @classmethod
    def get_block_receipts(
        self,
        block_number: Optional[int] = None,
        block_hash: Optional[HexStr] = None,
    ) -> RPCResponseModel[list[HexStr], list[TransactionReceiptModel]]:
        if block_number:
            param = hex(block_number)
        elif block_hash:
            param = block_hash
        return RPCResponseModel(
            self.rpc().get_block_receipts,
            [HexStr(param)],
        )

    @classmethod
    def alchemy_get_block_receipts(
        self,
        block_number: Optional[int] = None,
        block_hash: Optional[HexStr] = None,
    ) -> RPCResponseModel[AlchemyBlockReceipt, "AlchemyReceiptsResponse"]:
        return RPCResponseModel(
            self.rpc().alchemy_get_block_receipts,
            AlchemyBlockReceipt(
                params=AlchemyParams(
                    block_number=HexInteger(block_number) if block_number else None,
                    block_hash=block_hash,
                )
            ),
        )


class AlchemyReceiptsResponse(BaseModel):
    receipts: list[TransactionReceipt]


class Transaction(Request, TransactionModel, Generic[Network]):
    @classmethod
    def get_by_hash(
        self, tx_hash: HexStr
    ) -> RPCResponseModel[TransactionRequest, Optional["Transaction[Network]"]]:
        return RPCResponseModel(
            self.rpc().get_tx_by_hash,
            TransactionRequest(
                tx_hash=tx_hash,
            ),
        )

    @classmethod
    def get_pending_by_hash(
        self, tx_hash: HexStr
    ) -> RPCResponseModel[TransactionRequest, PendingTransaction]:
        return RPCResponseModel(
            self.rpc().get_pending_tx_by_hash,
            TransactionRequest(
                tx_hash=tx_hash,
            ),
        )

    @classmethod
    def get_receipt_by_hash(
        self, tx_hash: HexStr
    ) -> RPCResponseModel[TransactionRequest, "Transaction[Network]"]:
        return RPCResponseModel(
            self.rpc().get_tx_receipt,
            TransactionRequest(
                tx_hash=tx_hash,
            ),
        )

    @classmethod
    def send_raw_transaction(
        cls, tx: HexStr
    ) -> RPCResponseModel[RawTransaction, HexStr]:
        return RPCResponseModel(
            cls.rpc().send_raw_tx,
            RawTransaction(
                signed_tx=tx,
            ),
        )

    @classmethod
    def get_by_index(
        self,
        transaction_index: int,
        block_hash: HexStr | None = None,
        block_number: int | BLOCK_STRINGS | None = None,
    ) -> RPCResponseModel[
        GetTransactionByBlockHash | GetTransactionByBlockNumber, "Transaction[Network]"
    ]:
        if block_hash is None and block_number is None:
            raise ValueError("Must provide either block_hash or block_number")
        if block_hash:
            return RPCResponseModel(
                self.rpc().get_tx_by_block_hash,
                GetTransactionByBlockHash(
                    block_hash=block_hash,
                    index=HexInteger(transaction_index),
                ),
            )
        block_number = cast(int | BLOCK_STRINGS, block_number)
        return RPCResponseModel(
            self.rpc().get_tx_by_block_number,
            GetTransactionByBlockNumber(
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
                index=HexInteger(transaction_index),
            ),
        )

    @classmethod
    async def subscribe_pending(  # noqa: C901
        self,
    ) -> AsyncIterator[PendingTransaction]:  # noqa: C901
        rpc = _force_get_global_rpc()
        async for w3_connection in connect(
            rpc.wss,
            ping_interval=60,
            ping_timeout=60,
            max_queue=10000,
            open_timeout=30,
        ):
            try:
                await self._send_subscription_request(
                    w3_connection,
                )
                subscription_response: SubscriptionResponse = json.loads(
                    await w3_connection.recv()
                )
                if not subscription_response.get("result"):
                    raise ValueError(subscription_response)
            except Exception as e:
                raise e

            while True:
                try:
                    message = await asyncio.wait_for(w3_connection.recv(), timeout=32.0)
                    message_json: JsonPendingWssResponse = json.loads(message)
                    if "params" not in message_json:
                        raise ValueError(message_json)

                    transaction_hash: HexStr = message_json["params"]["result"]
                    transaction = await self.get_pending_by_hash(transaction_hash)
                    if transaction:
                        yield transaction
                except asyncio.exceptions.TimeoutError:
                    pass
                except (
                    ConnectionClosedError,
                    ConnectionResetError,
                    OSError,  # No route to host
                    asyncio.exceptions.IncompleteReadError,  # TODO: should this be handled differently?
                ) as err:
                    logger.error("connection terminated unexpectedly: %s", err)
                    await asyncio.sleep(1)
                    # we're in an iterator, so make a new connection and continue listening
                    break
                except Exception as err:
                    logger.error("unknown connection error: %s", err)
                    await asyncio.sleep(1)
                    # we're in an iterator, so make a new connection and continue listening
                    break

    @staticmethod
    async def _send_subscription_request(
        w3_connection: WebSocketClientProtocol,
    ):
        # this only handles a list of topic0s being subscribed together
        await w3_connection.send(
            json.dumps(
                {
                    "id": 1,
                    "method": "eth_subscribe",
                    "params": ["newPendingTransactions"],
                }
            )
        )

    def receipt(
        self,
    ) -> RPCResponseModel[TransactionRequest, Optional[TransactionReceipt]]:
        return RPCResponseModel(
            self.rpc().get_tx_receipt,
            TransactionRequest(
                tx_hash=self.hash,
            ),
        )



import os
from typing import Annotated

import pytest
from eth_rpc import Event, set_alchemy_key
from eth_rpc.networks import Arbitrum, Ethereum
from eth_rpc.types import Indexed, primitives
from pydantic import BaseModel


class TransferEventType(BaseModel):
    sender: Annotated[primitives.address, Indexed]
    recipient: Annotated[primitives.address, Indexed]
    amount: primitives.uint256


TransferEvent = Event[TransferEventType](name="Transfer")


@pytest.mark.unit
@pytest.mark.event
def test_event_network():
    assert TransferEvent[Ethereum]._network == Ethereum
    assert TransferEvent[Arbitrum]._network == Arbitrum
    swap_topic = TransferEvent.get_topic0
    assert (
        swap_topic
        == "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
    )

    eth_event = TransferEvent[Ethereum]
    assert eth_event._network == Ethereum
    arb_event = TransferEvent[Arbitrum]
    assert arb_event._network == Arbitrum

    # ensure previous instance isn't changed
    assert eth_event._network == Ethereum


@pytest.mark.asyncio(scope="session")
@pytest.mark.event
async def test_event_lookup():
    set_alchemy_key(os.environ["ALCHEMY_KEY"])
    eth_ = TransferEvent[Ethereum]
    logs = await eth_._get_logs(10_000_000, 10_000_001)
    assert len(logs) == 121



import asyncio
import json
import re
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from contextvars import ContextVar
from typing import Generic, Optional, TypeVar

from eth_rpc import Event, EventData, get_current_network
from eth_rpc.block import Block
from eth_rpc.log import Log
from eth_rpc.types import (
    BLOCK_STRINGS,
    EvmDataDict,
    JsonResponseWssResponse,
    LogsArgs,
    LogsParams,
    SubscriptionResponse,
)
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict, Field
from websockets.exceptions import ConnectionClosedError
from websockets.legacy.client import WebSocketClientProtocol, connect

from .._request import Request
from ..constants import DEFAULT_CONTEXT, DEFAULT_EVENT

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


class Receiver(ABC, Generic[T]):
    @abstractmethod
    async def put(self, event: EventData[T]) -> None: ...


class EventSubscriber(Request, Generic[U]):
    receivers: dict[HexStr, list[Receiver[U]]] = Field(default_factory=dict)
    events: list[Event] = Field(default_factory=list)
    step_size: int | None = Field(default=None)

    _start_block: Optional[int | BLOCK_STRINGS] = None
    _end_block: Optional[int | BLOCK_STRINGS] = None
    _addresses: Optional[list[HexAddress]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def network(self):
        return self._network or get_current_network()

    def get_topics(self):
        """The nested list makes it so any match to topic0 will be selected"""
        return [[e.get_topic0 for e in self.events]]

    async def _get_logs(
        self,
        start_block: int,
        end_block: int,
        addresses: list[HexAddress] = [],
    ) -> AsyncIterator[EventData[U]]:
        topic_dict = {event.get_topic0: event for event in self.events}
        cur_end = start_block + self.step_size - 1 if self.step_size else end_block

        while True:
            try:
                response = await self.rpc().get_logs(
                    LogsArgs(
                        params=LogsParams(
                            address=addresses,
                            from_block=start_block,
                            to_block=cur_end,
                            topics=self.get_topics(),
                        )
                    )
                )
            except ValueError as err:
                # TODO: confirm this error is due to the cur_end being too far in the future
                message = err.args[0]

                if "Log response size exceeded." in message:
                    boundaries = re.findall("0x[0-9a-f]+", message)
                    if len(boundaries) != 2:
                        raise err
                    cur_end = int(boundaries[1], 16)
                else:
                    raise err
                continue

            for result in response:
                event = topic_dict[result.topics[0]]

                # TODO: this is just a placeholder
                if len(result.topics) != (len(event.get_indexed()) + 1):
                    # this happens when an event has the same topic0, but different indexed events so it doesn't match up to the expected ABI
                    # print("INDEX MISMATCH", result.transaction_hash, result.log_index)
                    continue

                event_data = EventData[U](
                    name=event.name,
                    log=result,
                    event=event.process(
                        result.topics,
                        result.data,
                    ),
                    network=self.network,
                )
                yield event_data

            start_block = cur_end + 1
            cur_end = start_block + self.step_size - 1 if self.step_size else end_block
            if start_block >= end_block:
                break

    async def get_logs(
        self,
        start_block: int | BLOCK_STRINGS,
        end_block: int | BLOCK_STRINGS,
        addresses: list[HexAddress] = [],
    ):
        _start_block: int = await Block.convert(start_block)
        _end_block: int = await Block.convert(end_block)
        async for event_data in self._get_logs(_start_block, _end_block, addresses):
            for receiver in self.receivers[event_data.tx.topics[0]]:
                await receiver.put(event_data)

    def add_receiver(self, receiver: Receiver, events: list[Event[U]]):
        for event in events:
            # add event to events list
            if event not in self.events:
                self.events.append(event)

            # add topic0 to receivers
            topic0 = event.get_topic0
            if topic0 not in self.receivers:
                self.receivers[topic0] = [receiver]
            else:
                if receiver not in self.receivers[topic0]:
                    self.receivers[topic0].append(receiver)

    async def _get_event(
        self,
        w3_connection: WebSocketClientProtocol,
        event_dict: dict[HexStr, "Event"],
    ) -> EventData[U] | None:
        message = await asyncio.wait_for(w3_connection.recv(), timeout=32.0)
        message_json: JsonResponseWssResponse = json.loads(message)
        if "params" not in message_json:
            raise ValueError(message_json)

        result_dict: EvmDataDict = message_json["params"]["result"]
        if result_dict["removed"]:
            return None
        result = Log(**result_dict, network=self.network)  # type: ignore
        event = event_dict[result.topics[0]]

        # TODO: this is just a placeholder
        if len(result.topics) != (len(event.get_indexed()) + 1):
            # print("INDEX MISMATCH", result.transaction_hash, result.log_index)
            return None

        return EventData[U](
            name=event.name,
            log=result,
            event=event.process(
                result.topics,
                result.data,
            ),
            network=self.network,
        )

    async def _listen(  # noqa: C901
        self,
        addresses: list[HexAddress] = [],
    ):
        # TODO: sometimes the topics match, but the indexed fields are different
        topic_dict = {event.get_topic0: event for event in self.events}

        rpc = self._rpc()
        async for w3_connection in connect(
            rpc.wss,
            ping_interval=60,
            ping_timeout=60,
            max_queue=10000,
            open_timeout=30,
        ):
            try:
                await self._send_subscription_request(
                    w3_connection,
                    addresses,
                    [event.get_topic0 for event in self.events],
                )
                subscription_response: SubscriptionResponse = json.loads(
                    await w3_connection.recv()
                )
                if not subscription_response.get("result"):
                    raise ValueError(subscription_response)
            except Exception as e:
                raise e

            while True:
                try:
                    event_data = await self._get_event(w3_connection, topic_dict)
                    if event_data is None:
                        continue
                    yield event_data

                except asyncio.exceptions.TimeoutError:
                    pass
                except (
                    ConnectionClosedError,
                    ConnectionResetError,
                    OSError,  # No route to host
                    asyncio.exceptions.IncompleteReadError,  # TODO: should this be handled differently?
                ) as err:
                    print("Connection Error:", err)
                    await asyncio.sleep(3)
                    # we're in an iterator, so make a new connection and continue listening
                    break
                except Exception as e:
                    print("Unknown Error:", e)
                    await asyncio.sleep(3)
                    # we're in an iterator, so make a new connection and continue listening
                    break

    async def listen(
        self,
        addresses: list[HexAddress] = [],
        publish_events: asyncio.Event = DEFAULT_EVENT,
        start_context: ContextVar = DEFAULT_CONTEXT,
    ):
        """
        Listen for new blockchain events.  Publish Event is used to tell the listener to
        start publishing to receivers.  The default is for this variable to be set.
        The start_context is used to set a starting block number, so you can synchronize this
        with backfill.
        """
        queue: asyncio.Queue[EventData[U]] = asyncio.Queue()
        flush_queue: bool = True
        async for event_data in self._listen(addresses=addresses):
            if publish_events.is_set():
                if flush_queue:
                    while not queue.empty():
                        event_data = await queue.get()
                        if event_data.tx.block_number <= start_context.get():
                            continue
                        for receiver in self.receivers[event_data.tx.topics[0]]:
                            await receiver.put(event_data)
                    flush_queue = False
                for receiver in self.receivers[event_data.tx.topics[0]]:
                    await receiver.put(event_data)
            else:
                await queue.put(event_data)

    async def backfill_and_listen(
        self,
        start_block: int,
        addresses: list[HexAddress] = [],
    ):
        should_publish_events = asyncio.Event()
        latest_block: ContextVar[int] = ContextVar("latest_block")
        task = asyncio.create_task(
            self.listen(
                addresses=addresses,
                publish_events=should_publish_events,
                start_context=latest_block,
            )
        )
        latest = await Block.latest()
        latest_block.set(latest.number)
        await self.get_logs(start_block, latest.number, addresses)
        should_publish_events.set()
        return await task

    @staticmethod
    async def _send_subscription_request(
        w3_connection: WebSocketClientProtocol,
        addresses: list[HexAddress],
        topics: list[HexStr],
    ) -> None:
        # this only handles a list of topic0s being subscribed together
        await w3_connection.send(
            json.dumps(
                {
                    "id": 1,
                    "method": "eth_subscribe",
                    "params": [
                        "logs",
                        {
                            "address": addresses,
                            "topics": [[t for t in topics]],
                        },
                    ],
                }
            )
        )

    def __call__(
        self,
        start_block: int | BLOCK_STRINGS = 0,
        end_block: int | BLOCK_STRINGS = "latest",
        addresses: list[HexAddress] = [],
    ) -> AsyncIterator[EventData[U]]:
        """This is used to set these values so you can use the iterator"""

        self._start_block = start_block
        self._end_block = end_block
        self._addresses = addresses
        return self.__aiter__()

    async def __aiter__(self) -> AsyncIterator[EventData[U]]:
        if not (
            self._start_block is not None
            and self._end_block is not None
            and self._addresses is not None
        ):
            raise ValueError("Must set blocks first.")
        start_block = await Block.convert(self._start_block)
        end_block = await Block.convert(self._end_block)
        async for event in self._get_logs(start_block, end_block, self._addresses):
            yield event



from .base import ProtocolBase
from .contract import Contract
from .eth_response import EthResponse
from .func_signature import FuncSignature
from .function import ContractFunc

__all__ = [
    "Contract",
    "ContractFunc",
    "EthResponse",
    "FuncSignature",
    "ProtocolBase",
]



from typing import ClassVar, Optional

from eth_rpc._request import Request
from eth_rpc.types import Network
from eth_typing import HexAddress, HexStr
from pydantic import Field, PrivateAttr


class ContractT(Request):
    _network: ClassVar[type[Network] | None] = PrivateAttr(default=None)

    address: HexAddress
    code_override: Optional[HexStr] = Field(default=None)



import zlib
from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Optional

from eth_rpc.types import HexInteger
from eth_rpc.utils import convert_datetime_to_iso_8601
from eth_typing import HexAddress, HexStr
from pydantic import Field, PlainSerializer, field_validator

from ..utils import BloomFilter, RPCModel, load_datetime_string

if TYPE_CHECKING:
    from eth_rpc.transaction import Transaction


class Block(RPCModel):
    number: HexInteger
    hash: Optional[HexStr] = None
    transactions: list["Transaction"] | list[HexStr] = Field(default_factory=list)
    base_fee_per_gas: Optional[HexInteger] = None
    difficulty: HexInteger
    extra_data: HexStr
    gas_limit: HexInteger
    gas_used: HexInteger
    logs_bloom: HexStr
    miner: Optional[HexAddress] = None
    mix_hash: HexStr
    nonce: Optional[HexStr] = None
    parent_hash: HexStr
    receipts_root: HexStr
    sha3_uncles: HexStr
    size: HexInteger
    state_root: HexStr
    timestamp: Annotated[datetime, PlainSerializer(convert_datetime_to_iso_8601)]
    total_difficulty: Optional[HexInteger] = None
    transactions_root: HexStr
    uncles: list[HexStr] = Field(default_factory=list)

    normalize_timestamp = field_validator("timestamp", mode="before")(
        load_datetime_string
    )

    def has_log(self, topic: HexStr):
        t = bytes.fromhex(topic.replace("0x", ""))
        return t in BloomFilter(self.logs_bloom)

    def compress(self) -> bytes:
        return zlib.compress(self.model_dump_json().encode("utf-8"))

    def __repr__(self):
        return f"<Block number={self.number}>"

    __str__ = __repr__



from inspect import isclass
from types import GenericAlias
from typing import Annotated, Any, get_args, get_origin

from eth_abi import decode, encode
from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import BaseModel

from .basic import ALL_PRIMITIVES


class Struct(BaseModel):
    """
    This is a type used to denote a solidity struct.  It is necessary
    to set this so the encoder knows to treat it as a single struct,
    and not a list of fields.

    For example, (uint8,bytes32) will encode differently from ((uint8,bytes32))

    One is a field with two arguments, the other is a field with a single argument,
        that is a struct with two fields
    """

    @classmethod
    def to_tuple_type(cls):
        lst = cls.to_type_list()
        return f"({','.join(lst)})"

    @classmethod
    def to_type_list(cls):
        from eth_rpc.utils.types import transform_primitive

        lst = []
        for _, field in cls.model_fields.items():
            type_ = field.annotation
            lst.append(transform_primitive(type_))
        return lst

    @classmethod
    def convert(cls, arg):
        if isinstance(arg, list):
            return [cls.convert(a) for a in arg]
        elif isinstance(arg, tuple):
            return tuple(cls.convert(a) for a in arg)
        elif isinstance(arg, Struct):
            return tuple(cls.convert(getattr(arg, a)) for a in arg.model_fields)
        return arg

    def to_bytes(self):
        types = self.to_tuple_type()
        values = []
        for _, value in self:
            values.append(self.convert(value))
        return encode([types], [tuple(values)])

    @classmethod
    def cast(cls, type_, args):
        if get_origin(type_) == list:
            (type_,) = get_args(type_)
            return [cls.cast(type_, arg) for arg in args]
        elif get_origin(type_) == tuple:
            tuple_types = get_args(type_)
            return tuple(cls.cast(type_, arg) for type_, arg in zip(tuple_types, args))
        elif isclass(type_) and issubclass(type_, Struct):
            return type_._build(args)
        return args

    @classmethod
    def from_bytes(cls, data_: bytes | HexStr):
        if isinstance(data_, str):
            bytes_ = bytes.fromhex(data_.removeprefix("0x"))
        else:
            bytes_ = data_
        types = cls.to_tuple_type()
        decoded = decode([types], bytes_)[0]
        return cls._build(decoded)

    @classmethod
    def from_tuple(cls, data: list[Any]):
        return cls._build(data)

    @classmethod
    def _build(cls, data):
        zipped = dict(zip(cls.model_fields.keys(), data))
        fields = [field.annotation for field in cls.model_fields.values()]
        response = {}
        for field, (name, value) in zip(fields, zipped.items()):
            response[name] = cls.cast(field, value)
        return cls(**response)

    @classmethod
    def struct_name(cls) -> str:
        """Override this if you want the struct name to be different from the class name"""
        return cls.__name__

    @staticmethod
    def transform(type_):
        mapping = {
            str: "string",
            int: "uint256",  # defaults to uint256
            HexAddress: "address",
            ChecksumAddress: "address",
            HexStr: "bytes",
        }
        if type_ in mapping:
            return mapping[type_]
        if str(type_) in ALL_PRIMITIVES:
            return str(type_)
        # handle list type
        if isinstance(type_, GenericAlias):
            (arg,) = get_args(type_)
            return f"{arg.__name__}[]"
        return type_.__name__

    @classmethod
    def get_nested_types(cls):
        for _, field in cls.model_fields.items():
            for item in cls._get_nested_types(field.annotation):
                yield item

    @classmethod
    def _get_nested_types(cls, type_) -> list:
        if get_origin(type_) == list:
            return cls._get_nested_types(get_args(type_)[0])
        elif get_origin(type_) == tuple:
            args = get_args(type_)
            return [
                item
                for subl in [cls._get_nested_types(arg) for arg in args]
                for item in subl
            ]
        elif get_origin(type_) == Annotated:
            return cls._get_nested_types(get_args(type_)[0])
        return []

    @classmethod
    def type_string(cls) -> bytes:
        type_string: str = ""
        nested_types = list(cls.get_nested_types())
        for name, field_data in cls.model_fields.items():
            name = field_data.serialization_alias or name
            type_ = field_data.annotation

            if isinstance(type_, GenericAlias):
                if get_origin(type_) == list:
                    (arg,) = get_args(type_)
                    type_str = f"{cls.transform(arg)}[]"
                elif get_origin(type_) == tuple:
                    args = get_args(type_)
                    type_str = f"({','.join([cls.transform(arg) for arg in args])})"
            else:
                type_str = cls.transform(type_)
            type_string += f"{type_str},"
        encoded_typestr: bytes = (
            f'{cls.struct_name()}({type_string.rstrip(",")})'.encode("utf-8")
        )
        return encoded_typestr + b"".join(set(nested_types))



from eth_rpc.types import HexInteger
from eth_rpc.utils import RPCModel
from eth_typing import HexStr


class Account(RPCModel):
    code_hash: HexStr
    storage_root: HexStr
    balance: HexInteger
    nonce: HexInteger



import secrets
from abc import ABC, abstractmethod
from typing import Any, Literal, Optional

from eth_account import Account as EthAccount
from eth_account.account import LocalAccount, SignedMessage
from eth_rpc.types import (
    BLOCK_STRINGS,
    CallWithBlockArgs,
    EthCallParams,
    GetAccountArgs,
    RawTransaction,
    SignedTransaction,
)
from eth_typing import HexAddress, HexStr
from pydantic import ConfigDict, PrivateAttr

from ._request import Request
from ._transport import _force_get_global_rpc
from .account import Account
from .block import Block
from .transaction import PreparedTransaction
from .types import HexInteger, RPCResponseModel


class BaseWallet(Request, ABC):
    @property
    @abstractmethod
    def address(self) -> HexAddress: ...

    def get_nonce(self, block_number: int | BLOCK_STRINGS = "latest"):
        return RPCResponseModel(
            self.rpc().get_tx_count,
            GetAccountArgs(
                address=self.address,
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
            ),
        )

    @abstractmethod
    def sign_transaction(self, tx) -> SignedTransaction: ...

    @abstractmethod
    def send_raw_transaction(
        self, tx: HexStr
    ) -> RPCResponseModel[RawTransaction, HexStr]: ...


class MockWallet(BaseWallet):
    _address: HexAddress = PrivateAttr()

    @property
    def address(self) -> HexAddress:
        return self._address

    def sign_transaction(self, tx):
        raise NotImplementedError("Mock wallet can not sign")


class PrivateKeyWallet(BaseWallet):
    private_key: HexStr
    _account: LocalAccount = PrivateAttr()

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context: Any) -> None:
        self._account = EthAccount.from_key(self.private_key)
        return super().model_post_init(__context)

    @property
    def address(self) -> HexAddress:
        return self._account.address

    @staticmethod
    def get_pvt_key() -> HexStr:
        priv = secrets.token_hex(32)
        return HexStr(f"0x{priv}")

    @classmethod
    def create_new(cls):
        return cls(private_key=cls.get_pvt_key())

    def sign_transaction(self, tx: PreparedTransaction) -> SignedTransaction:
        signed_tx = self._account.sign_transaction(tx.model_dump())
        return SignedTransaction(
            raw_transaction=signed_tx[0].hex(),
            hash=signed_tx.hash.hex(),
            r=signed_tx.r,
            s=signed_tx.s,
            v=signed_tx.v,
        )

    def send_raw_transaction(
        self, tx: HexStr
    ) -> RPCResponseModel[RawTransaction, HexStr]:
        return RPCResponseModel(
            self.rpc().send_raw_tx,
            RawTransaction(
                signed_tx=tx,
            ),
        )

    def prepare_and_sign(
        self,
        *,
        to: HexAddress,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        data: HexStr = HexStr("0x"),
        nonce: Optional[int] = None,
    ):
        prepared = self.prepare(
            to=to,
            value=value,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            data=data,
            nonce=nonce,
        )
        return self.sign_transaction(prepared)

    def prepare(
        self,
        *,
        to: HexAddress,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        data: HexStr = HexStr("0x"),
        nonce: Optional[int] = None,
    ):
        # TODO: this assumes sync
        gas = self.estimate_gas(to=to, data=data).sync
        access_list = None
        rpc = _force_get_global_rpc()
        chain_id = rpc.chain_id.sync()

        max_priority_fee_per_gas = max_priority_fee_per_gas or Block.priority_fee().sync
        base_fee_per_gas = Block.pending().sync.base_fee_per_gas
        assert base_fee_per_gas, "block is earlier than London Hard Fork"
        max_fee_per_gas = max_fee_per_gas or (
            2 * base_fee_per_gas + max_priority_fee_per_gas
        )

        return PreparedTransaction(
            data=data,
            to=to,
            gas=HexInteger(gas),
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            nonce=nonce or self.get_nonce().sync,
            value=value,
            access_list=access_list,
            chain_id=chain_id,
        )

    def estimate_gas(
        self,
        to: HexAddress,
        block_number: HexInteger | Literal["latest", "pending"] = "latest",
        data: HexStr = HexStr("0x"),
    ) -> RPCResponseModel[CallWithBlockArgs, HexInteger]:
        return RPCResponseModel(
            self.rpc().estimate_gas,
            CallWithBlockArgs(
                params=EthCallParams(
                    from_=self.address,
                    to=to,
                    data=data,
                ),
                block_number=block_number,
            ),
        )

    def transfer(
        self, to: HexAddress, value: int
    ) -> RPCResponseModel[RawTransaction, HexStr]:
        prepared_tx = self.prepare(to=to, value=value)
        signed_tx = self.sign_transaction(prepared_tx)
        return self.send_raw_transaction(HexStr("0x" + signed_tx.raw_transaction))

    def sign_hash(self, hashed: bytes) -> SignedMessage:
        return EthAccount._sign_hash(hashed, self._account.key)

    async def balance(self, block_number: int | BLOCK_STRINGS = "latest") -> int:
        return await Account.get_balance(self.address, block_number=block_number)

    @staticmethod
    def rsv_to_signature(r: int, s: int, v: int) -> HexStr:
        rr = hex(r)[2:].zfill(64)
        ss = hex(s)[2:].zfill(64)
        vv = hex(v)[2:]
        return HexStr("0x" + rr + ss + vv)

    @staticmethod
    def signature_to_rsv(signature: HexStr) -> tuple[int, int, int]:
        v = signature[-2:]
        s = signature[-66:-2]
        r = signature[:-66].lstrip("0x")

        return int(r, 16), int(s, 16), int(v, 16)



import asyncio
import json
import logging
import re
from copy import deepcopy
from functools import cached_property
from types import GenericAlias
from typing import Any, AsyncIterator, Generic, Literal, Optional, TypeVar, get_args

from eth_abi import decode
from eth_abi.exceptions import InsufficientDataBytes
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, PrivateAttr, computed_field
from pydantic_core import Url
from websockets.exceptions import ConnectionClosedError
from websockets.legacy.client import WebSocketClientProtocol, connect

from ._request import Request
from ._transport import _force_get_default_network, get_current_network
from .block import Block
from .exceptions import LogDecodeError, LogResponseExceededError, RateLimitingError
from .models import EventData, Log
from .types import (
    BLOCK_STRINGS,
    BlockReference,
    EvmDataDict,
    Indexed,
    JsonResponseWssResponse,
    LogsArgs,
    LogsParams,
    Name,
    Network,
    RPCResponseModel,
    SubscriptionResponse,
)
from .utils import is_annotation, to_topic

T = TypeVar("T", bound=BaseModel)
logger = logging.getLogger(__name__)
IGNORE = Literal[""]
IGNORE_VAL: IGNORE = ""


def map_type(name):
    return {
        HexAddress: "address",
    }.get(name, name.__name__)


def map_indexed(is_indexed):
    if is_indexed:
        return " indexed "
    return ""


def convert(type, with_name: bool = False, with_indexed: bool = False):
    name = ""
    indexed = False
    if is_annotation(type):
        type, *annotations = get_args(type)
        for annotation in annotations:
            if annotation == Indexed:
                if with_indexed:
                    indexed = True
            elif isinstance(annotation, Name):
                if with_name:
                    name = annotation.value
    if isinstance(type, GenericAlias):
        if type.__name__ == "list":
            list_type = get_args(type)[0]
            converted_list_type = convert(list_type)
            if isinstance(converted_list_type, list):
                converted_list_type = f"({','.join(converted_list_type)})"
            return f"{converted_list_type}[] {name}".strip()
        else:
            tuple_args = ",".join([convert(t) for t in get_args(type)])
            return f"({tuple_args})"
    return f"{map_type(type)}{map_indexed(indexed)}{name}".strip()


class Event(Request, Generic[T]):
    name: str
    anonymous: bool = False

    topic1_filter: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL
    topic2_filter: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL
    topic3_filter: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL
    addresses_filter: list[HexAddress] = []

    _output_type: BaseModel = PrivateAttr()

    def model_post_init(self, __context) -> None:
        EventType, *_ = self.__pydantic_generic_metadata__["args"]
        self._output_type = EventType
        return super().model_post_init(__context)

    @staticmethod
    def _matches(topic: HexStr, topic_filter: HexStr | list[HexStr] | None) -> bool:
        if isinstance(topic_filter, list):
            if None in topic_filter:
                return True
            return topic in topic_filter
        return topic == topic_filter

    def match_topics(self, log: Log) -> bool:
        # TODO: addresses_filter
        if len(log.topics) == 0:
            return False
        if log.topics[0] != self.get_topic0:
            return False
        if self.topic1_filter != IGNORE_VAL and len(log.topics) >= 2:
            if not self._matches(log.topics[1], self.topic1_filter):
                return False
        if self.topic2_filter != IGNORE_VAL and len(log.topics) >= 3:
            if not self._matches(log.topics[2], self.topic2_filter):
                return False
        if self.topic3_filter != IGNORE_VAL and len(log.topics) >= 4:
            if not self._matches(log.topics[3], self.topic3_filter):
                return False
        return True

    def match_address(self, log: Log) -> bool:
        if self.addresses_filter:
            return log.address in self.addresses_filter
        return True

    def match(self, log: Log) -> bool:
        return self.match_address(log) and self.match_topics(log)

    def add_address(self, address: HexAddress | list[HexAddress]):
        if isinstance(address, list):
            for addr in address:
                self.add_address(addr)
        else:
            if address not in self.addresses_filter:
                self.addresses_filter.append(address)

    def remove_address(self, address: HexAddress):
        self.addresses_filter = [
            _address for _address in self.addresses_filter if _address != address
        ]

    @staticmethod
    def process_value(type_name, v: str):
        # strip prefix if necessary
        if "0x" in v:
            v = v[2:]

        if type_name == "address":
            # last 20 bytes of value
            return "0x{}".format(v[-40:])
        if "bytes" in type_name:
            return bytes.fromhex(v)
        if "uint" in type_name:
            return int.from_bytes(bytes.fromhex(v), "big", signed=False)
        elif "int" in type_name:
            return int.from_bytes(bytes.fromhex(v), "big", signed=True)
        if type_name == "bool":
            return v[-1] == "1"

    def from_dict(self, fields: dict[str, Any]):
        EventType, *_ = self.__pydantic_generic_metadata__["args"]
        return EventType(**fields)

    def process_log(self, log: Log) -> EventData[T]:
        return EventData(
            name=self.name,
            log=log,
            event=self.process(log.topics, log.data),
            network=self._network or get_current_network(),
        )

    def process(self, topics: list[HexStr], data: HexStr) -> T:
        EventType, *_ = self.__pydantic_generic_metadata__["args"]
        indexed = self.get_indexed()
        try:
            indexed_dict = {
                name: self.process_value(type_, topics[i + 1])
                for i, (name, type_) in enumerate(indexed)
            }
        except IndexError:
            raise LogDecodeError("Mismatched Indexed values")

        unindexed = self.get_unindexed()
        try:
            unindexed_values = decode(
                [type_ for (_, type_) in unindexed], bytes.fromhex(data[2:])
            )
        except InsufficientDataBytes:
            raise LogDecodeError("Mismatched Unindexed values")

        return EventType(
            **indexed_dict
            | {name: val for (name, _), val in zip(unindexed, unindexed_values)}
        )

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def get_topic0(self) -> HexStr:
        inputs, *_ = self.__pydantic_generic_metadata__["args"]
        input_types = inputs.model_fields
        converted_inputs = ",".join(
            [convert(field.annotation) for name, field in input_types.items()]
        )
        event_signature = f"{self.name}({converted_inputs})"
        return HexStr("0x" + to_topic(event_signature).hex())

    def get_indexed(self):
        inputs, *_ = self.__pydantic_generic_metadata__["args"]
        input_types = inputs.model_fields
        results = []
        for name, field in input_types.items():
            indexed = False
            _type = field.annotation
            annotations = field.metadata
            for annotation in annotations:
                if annotation == Indexed:
                    indexed = True
                elif isinstance(annotation, Name):
                    name = annotation.value
            if indexed:
                results.append((name, map_type(_type)))
        return results

    def get_unindexed(self):
        inputs, *_ = self.__pydantic_generic_metadata__["args"]
        input_types = inputs.model_fields
        results = []
        for name, field in input_types.items():
            indexed = False
            _type = field.annotation
            annotations = field.metadata
            for annotation in annotations:
                if annotation == Indexed:
                    indexed = True
                elif isinstance(annotation, Name):
                    name = annotation.value
            if not indexed:
                results.append((name, convert(_type)))
        return results

    def _get_logs(
        self,
        start_block: BlockReference | int,
        end_block: BlockReference | int,
        addresses: list[HexAddress] = [],
        topic1: Optional[HexStr | list[HexStr] | None] | IGNORE = IGNORE_VAL,
        topic2: Optional[HexStr | list[HexStr] | None] | IGNORE = IGNORE_VAL,
        topic3: Optional[HexStr | list[HexStr] | None] | IGNORE = IGNORE_VAL,
    ) -> RPCResponseModel[LogsArgs, list[Log]]:
        topics: list[HexStr | None | list[HexStr]] = []
        if topic1 != IGNORE_VAL:
            topics.append(topic1)
            if topic2 != IGNORE_VAL:
                topics.append(topic2)
                if topic3 != IGNORE_VAL:
                    topics.append(topic3)
        return RPCResponseModel(
            self.rpc().get_logs,
            LogsArgs(
                params=LogsParams(
                    address=addresses,
                    from_block=start_block,
                    to_block=end_block,
                    topics=[self.get_topic0, *topics],
                )
            ),
        )

    def set_filter(
        self,
        addresses: list[HexAddress] = [],
        topic1: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL,
        topic2: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL,
        topic3: Optional[HexStr | list[HexStr]] | IGNORE = IGNORE_VAL,
    ):
        model = deepcopy(self)
        model.addresses_filter = addresses
        if topic1 != IGNORE_VAL:
            model.topic1_filter = topic1
        if topic2 != IGNORE_VAL:
            model.topic1_filter = model.topic1_filter or None
            model.topic2_filter = topic2
        if topic3 != IGNORE_VAL:
            model.topic1_filter = model.topic1_filter or None
            model.topic2_filter = model.topic2_filter or None
            model.topic3_filter = topic3
        return model

    async def get_logs(
        self,
        start_block: BlockReference | int,
        end_block: BlockReference | int,
    ) -> AsyncIterator[EventData[T]]:
        cur_end = end_block
        try:
            response = await self._get_logs(
                start_block,
                cur_end,
                self.addresses_filter,
                topic1=self.topic1_filter,
                topic2=self.topic2_filter,
                topic3=self.topic3_filter,
            )
        except ValueError as err:
            message = err.args[0]
            if "Log response size exceeded." in message:
                boundaries = re.findall("0x[0-9a-f]+", message)
                raise LogResponseExceededError(
                    err.args[0], int(boundaries[0], 16), int(boundaries[1], 16)
                )
            elif (
                "Your app has exceeded its compute units per second capacity" in message
            ):
                raise RateLimitingError(message)
            raise err

        for result in response:
            # TODO: this is just a placeholder
            if len(result.topics) != (len(self.get_indexed()) + 1):
                # this happens when an event has the same topic0, but different indexed events so it doesn't match up to the expected ABI
                continue

            event_data = EventData[T](
                name=self.name,
                log=result,
                event=self.process(
                    result.topics,
                    result.data,
                ),
                network=self._network or get_current_network(),
            )
            yield event_data

    async def backfill(
        self,
        start_block: int | None = None,
        end_block: int | None = None,
        step_size: Optional[int] = None,
    ) -> AsyncIterator[EventData[T]]:
        """
        This backfills events, handling LogResponseExceededError to provide all logs in a range too large for a single request
        """
        start_block = start_block or 1
        current_number = await Block.get_number()
        end_block = end_block or (current_number - 3)  # set 3 default confirmations

        if start_block == "earliest":
            cur_start = 0
        else:
            cur_start = start_block

        if step_size:
            cur_end = cur_start + step_size
        else:
            cur_end = end_block
        while cur_start <= end_block:
            try:
                async for log in self.get_logs(
                    start_block=cur_start,
                    end_block=min(cur_end, end_block),
                ):
                    yield log
            except LogResponseExceededError as err:
                cur_end = err.recommended_end
                continue
            except RateLimitingError:
                await asyncio.sleep(3)
                continue
            cur_start = cur_end + 1
            if step_size:
                cur_end += step_size
            else:
                cur_end = end_block

    @property
    def subscribe(self) -> "AsyncSubscribeCallable[T]":
        """
        This returns a callable for the async subscriber, allowing you to select the network
        for the subscription, ie.

        ```python
        my_event = Event[EventType](name="MyEvent")
        async for event in my_event.subscribe[Ethereum]():
            ...
        ```

        If no network is provided, it will use the default network.
        """
        # TODO: sometimes the topics match, but the indexed fields are different
        network = self._network or _force_get_default_network()
        return AsyncSubscribeCallable(
            network=network,
            event=self,
        )

    @staticmethod
    async def _convert_block_string(
        block_string: BLOCK_STRINGS,
    ) -> int:
        if block_string == "earliest":
            return 0
        return (await Block.load_by_number(block_string)).number

    @staticmethod
    async def _send_subscription_request(
        w3_connection: WebSocketClientProtocol,
        pending: bool,
        addresses: list[HexAddress],
        topics: list[HexStr | list[HexStr] | None],
    ):
        # this only handles a list of topic0s being subscribed together
        subscription_type = "newPendingTransactions" if pending else "logs"
        await w3_connection.send(
            json.dumps(
                {
                    "id": 1,
                    "method": "eth_subscribe",
                    "params": [
                        subscription_type,
                        {
                            "address": addresses,
                            "topics": topics,
                        },
                    ],
                }
            )
        )


class AsyncSubscribeCallable(BaseModel, Generic[T]):
    network: type[Network]
    event: Event[T]

    def __getitem__(self, network: type[Network]):
        self.network = network
        return self

    async def __call__(  # noqa: C901
        self,
    ) -> AsyncIterator[EventData[T]]:
        # TODO: sometimes the topics match, but the indexed fields are different

        if not (wss_uri := self.network.wss):
            raise ValueError("No wss set for network")

        async for w3_connection in connect(
            wss_uri.unicode_string() if isinstance(wss_uri, Url) else wss_uri,
            ping_interval=60,
            ping_timeout=60,
            max_queue=10000,
            open_timeout=30,
        ):
            try:
                topics: list[HexStr | list[HexStr] | None] = []
                if self.event.topic1_filter != IGNORE_VAL:
                    topics.append(self.event.topic1_filter)
                    if self.event.topic2_filter != IGNORE_VAL:
                        topics.append(self.event.topic2_filter)
                        if self.event.topic3_filter != IGNORE_VAL:
                            topics.append(self.event.topic3_filter)

                await self.event._send_subscription_request(
                    w3_connection,
                    False,
                    self.event.addresses_filter,
                    [
                        self.event.get_topic0,
                        *topics,
                    ],
                )
                subscription_response: SubscriptionResponse = json.loads(
                    await w3_connection.recv()
                )
                if not subscription_response.get("result"):
                    raise ValueError(subscription_response)
            except Exception as e:
                raise e

            while True:
                try:
                    message = await asyncio.wait_for(w3_connection.recv(), timeout=32.0)
                    message_json: JsonResponseWssResponse = json.loads(message)
                    if "params" not in message_json:
                        raise ValueError(message_json)

                    result_dict: EvmDataDict = message_json["params"]["result"]
                    result = Log.model_validate(result_dict)
                    yield EventData(
                        name=self.event.name,
                        log=result,
                        event=self.event.process(
                            result.topics,
                            result.data,
                        ),
                        network=self.network,
                    )
                except asyncio.exceptions.TimeoutError:
                    pass
                except (TypeError, LogDecodeError):
                    logger.warning("Mistmatched type: %s", result_dict)
                except (
                    ConnectionClosedError,
                    ConnectionResetError,
                    OSError,  # No route to host
                    asyncio.exceptions.IncompleteReadError,  # TODO: should this be handled differently?
                ) as err:
                    logger.error("connection terminated unexpectedly: %s", err)
                    await asyncio.sleep(1)
                    # we're in an iterator, so make a new connection and continue listening
                    break
                except Exception as err:
                    logger.error("connection error: %s", err)
                    await asyncio.sleep(1)
                    # we're in an iterator, so make a new connection and continue listening
                    break



from eth_typing import HexStr
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

from ..basic import BLOCK_STRINGS, HexInteger


class GetBlockByHashArgs(BaseModel):
    block_hash: HexStr
    with_tx_data: bool = False


class BlockNumberArg(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )
    block_number: HexInteger | BLOCK_STRINGS


class GetBlockByNumberArgs(BlockNumberArg):
    with_tx_data: bool = False



import os
from typing import Annotated

import pytest
from eth_abi import encode
from eth_rpc import ContractFunc, FuncSignature, set_alchemy_key
from eth_rpc.contract import EthResponse, ProtocolBase
from eth_rpc.networks import Arbitrum, Ethereum
from eth_rpc.types import METHOD, Name, NoArgs, Struct, primitives
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel


class AllowanceRequest(BaseModel):
    owner: primitives.address
    spender: primitives.address


class Token(ProtocolBase):
    name: ContractFunc[
        NoArgs,
        Annotated[primitives.string, Name("_name")],
    ] = METHOD
    balance_of: Annotated[
        ContractFunc[
            primitives.address,
            Annotated[primitives.uint256, Name("_name")],
        ],
        Name("balanceOf"),
    ] = METHOD
    allowance: ContractFunc[
        tuple[primitives.address, primitives.address],
        primitives.uint256,
    ] = METHOD
    allowance2: Annotated[
        ContractFunc[
            AllowanceRequest,
            primitives.uint256,
        ],
        Name("allowance"),
    ] = METHOD


@pytest.mark.contract
@pytest.mark.unit
def test_encode_signature():
    func = FuncSignature[NoArgs, primitives.string](name="name")
    assert func.encode_call(inputs=()) == "0x06fdde03"


@pytest.mark.unit
def test_set_network():
    usdt = Token[Ethereum](
        address=HexAddress(HexStr("0xdAC17F958D2ee523a2206206994597C13D831ec7"))
    )
    assert usdt._network == Ethereum

    usdc = Token[Arbitrum](address="0xaf88d065e77c8cc2239327c5edb3a432268e5831")
    assert usdc._network == Arbitrum

    # test that this didn't affect the usdt network
    assert usdt._network == Ethereum


@pytest.mark.asyncio(scope="session")
@pytest.mark.contract
async def test_contract() -> None:
    set_alchemy_key(os.environ["ALCHEMY_KEY"])

    usdt = Token[Ethereum](
        address=HexAddress(HexStr("0xdAC17F958D2ee523a2206206994597C13D831ec7"))
    )
    usdc = Token[Arbitrum](address="0xaf88d065e77c8cc2239327c5edb3a432268e5831")
    assert usdt._network == Ethereum
    assert usdc._network == Arbitrum

    assert await usdt.name().get() == "Tether USD"
    assert await usdc.name().get() == "USD Coin"

    binance_balance = await usdt.balance_of(
        "0xF977814e90dA44bFA03b6295A0616a897441aceC"
    ).get()
    binance_balance = await usdt.balance_of(
        "0xF977814e90dA44bFA03b6295A0616a897441aceC"
    ).get(block_number=20_604_386)
    assert binance_balance == 6600822508869000

    # check function with two args
    allowance = await usdt.allowance(
        "0x71AC0c3E825f095E56b81fBA11020f01893d4433",
        "0x38C5412464A03EfDc3820d227b24316C11729E0a",
    ).get(block_number=20604634)
    assert allowance == 126967536
    # check function using BaseModel for args
    allowance = await usdt.allowance2(
        "0x71AC0c3E825f095E56b81fBA11020f01893d4433",
        "0x38C5412464A03EfDc3820d227b24316C11729E0a",
    ).get(block_number=20604640)
    assert allowance == 0


@pytest.mark.unit
def test_multi_arg_func() -> None:
    response = EthResponse(
        func=FuncSignature[
            tuple[primitives.bytes32, primitives.bytes32, int],
            primitives.address,
        ](name="computeAddress"),
        result=HexStr(
            "0x0000000000000000000000008284ff4a6881edfa14435e1b140c9430e3a0ec5d"
        ),
    )

    assert response.decode() == "0x8284ff4a6881edfa14435e1b140c9430e3a0ec5d"


@pytest.mark.unit
def test_decode_struct() -> None:
    class SubStruct(Struct):
        x: bool
        y: list[primitives.uint256]

    class MyStruct(Struct):
        a: list[primitives.uint256]
        b: str
        c: list[SubStruct]

    my_struct = MyStruct(a=[1, 2, 3], b="test", c=[SubStruct(x=True, y=[1, 2, 3])])

    bytes_ = encode(
        ("(uint256[],string,(bool,uint256[])[])",),
        [([1, 2, 3], "test", [(True, [1, 2, 3])])],
    )
    assert my_struct.to_bytes().hex() == bytes_.hex()

    response = EthResponse(
        func=FuncSignature[
            NoArgs,
            MyStruct,
        ](name="computeAddress"),
        result=my_struct.to_bytes().hex(),
    )
    decoded = response.decode()

    assert isinstance(decoded, MyStruct)
    assert decoded.a == [1, 2, 3]
    assert decoded.b == "test"
    assert decoded.c[0].x
    assert decoded.c[0].y == [1, 2, 3]


@pytest.mark.unit
def test_decode_basemodel() -> None:
    class SubStruct(Struct):
        x: bool
        y: list[primitives.uint256]

    class MyStruct(BaseModel):
        a: list[primitives.uint256]
        b: str
        c: list[SubStruct]

    my_struct = MyStruct(a=[1, 2, 3], b="test", c=[SubStruct(x=True, y=[1, 2, 3])])
    bytes_ = encode(
        (
            "uint256[]",
            "string",
            "(bool,uint256[])[]",
        ),
        [[1, 2, 3], "test", [(True, [1, 2, 3])]],
    )

    response = EthResponse(
        func=FuncSignature[
            NoArgs,
            MyStruct,
        ](name="computeAddress"),
        result=HexStr(bytes_.hex()),
    )
    decoded = response.decode()
    assert decoded == my_struct



import asyncio
from contextvars import ContextVar

from eth_typing import HexAddress, HexStr

ADDRESS_ZERO = HexAddress(HexStr("0x0000000000000000000000000000000000000000"))
ADDRESS_DEAD = HexAddress(HexStr("0x000000000000000000000000000000000000dEaD"))

DEFAULT_EVENT = asyncio.Event()
DEFAULT_EVENT.set()
DEFAULT_CONTEXT = ContextVar[int]("DEFAULT_CONTEXT", default=0)



from typing import TYPE_CHECKING, Generator, Generic, NewType, Optional, TypeVar

if TYPE_CHECKING:
    from eth_rpc.rpc.method import RPCMethod

# used to indicate methods with no arguments
NoArgs = NewType("NoArgs", tuple[()])

ArgType = TypeVar("ArgType")
ReturnType = TypeVar("ReturnType")


class RPCResponseModel(Generic[ArgType, ReturnType]):
    __slots__ = ["func", "arg"]

    func: "RPCMethod"
    arg: ArgType | NoArgs

    def __init__(self, func: "RPCMethod", arg: Optional[ArgType] = None):
        self.func = func
        self.arg = arg if arg is not None else NoArgs(())

    @property
    def sync(self) -> ReturnType:
        if self.arg:
            return self.func.sync(self.arg)
        response = self.func.sync()
        return response

    def __await__(self) -> Generator[None, None, ReturnType]:
        if self.arg:
            return self.func(self.arg).__await__()
        return self.func().__await__()



from typing import ClassVar

from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url


class Ethereum(Network):
    chain_id: ClassVar[int] = 1
    name: ClassVar[str] = "Ethereum"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://cloudflare-eth.com"),
            wss=Url("wss://mainnet.gateway.tenderly.co"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Etherscan",
        url="https://etherscan.io",
        api_url="https://api.etherscan.io/api",
    )
    alchemy_str: ClassVar[str | None] = "eth-mainnet"
    apprx_block_time: ClassVar[float] = 12.05



from typing import Self

from eth_abi import decode
from eth_hash.auto import keccak
from eth_typing import HexStr

from .struct import Struct


class EvmError(Struct):
    @classmethod
    def selector(cls):
        return f"0x{keccak(cls.type_string())[:4].hex()}"

    @classmethod
    def decode(cls, data: HexStr) -> Self:
        selector = cls.selector()
        if not data.startswith(selector):
            raise ValueError(f"Invalid selector for {cls.__name__}")
        data = HexStr(data.removeprefix(selector))
        tuple_type = cls.to_tuple_type()
        decoded = decode([tuple_type], bytes.fromhex(data))[0]
        return cls.from_tuple(decoded)



from typing import ClassVar

from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url


class Arbitrum(Network):
    chain_id: ClassVar[int] = 42161
    name: ClassVar[str] = "Arbitrum"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://arb1.arbitrum.io/rpc"),
            wss=None,
            # wss=Url("wss://test.com"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Arbiscan",
        url="https://arbiscan.io",
        api_url="https://api.arbiscan.io/api",
    )
    alchemy_str: ClassVar[str | None] = "arb-mainnet"
    apprx_block_time: ClassVar[float] = 0.26


class ArbitrumSepolia(Network):
    chain_id: ClassVar[int] = 421614
    name: ClassVar[str] = "ArbitrumSepolia"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://arbitrum-sepolia.publicnode.com"),
            wss=None,
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Arbiscan",
        url="https://sepolia.arbiscan.io",
        api_url="https://api-sepolia.arbiscan.io/api",
    )
    alchemy_str: ClassVar[str | None] = "arb-sepolia"
    apprx_block_time: ClassVar[float] = 0.26



import itertools

import httpx
from eth_rpc.types import Network
from pydantic import BaseModel, ConfigDict, Field


class BaseRPC(BaseModel):
    _timeout: float = 10.0
    _retries: int = 3

    network: type[Network]

    index: itertools.count = Field(default_factory=lambda: itertools.count())
    client: httpx.AsyncClient = Field(default_factory=lambda: httpx.AsyncClient())

    @property
    def timeout(self) -> httpx.Timeout:
        """Request Timeout"""
        return httpx.Timeout(self._timeout)

    def set_timeout(self, timeout: float):
        self._timeout = timeout

    def set_retries(self, retries: int):
        self._retries = retries

    @property
    def retries(self):
        return self._retries

    @property
    def wss(self) -> str:
        if (wss := self.network.wss) is None:
            raise ValueError("wss not set")
        return str(wss)

    @property
    def http(self) -> str:
        if (http := self.network.http) is None:
            raise ValueError("http not set")
        return str(http)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )
